#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2013-2019

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Leon Majewski  <leon.majewski@bom.gov.au>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from libc.math cimport fabs, isinf, isnan
from libc.stdio cimport printf

import numpy as np

cimport numpy as np

DTYPE = np.double
ctypedef np.double_t DTYPE_t
cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void nn(const DTYPE_t[:, :, :] data, int l0, int p0, double dl, double dp, int lmax, int pmax, DTYPE_t[:] res) nogil:
    cdef int nnl, nnp
    cdef size_t z_size = res.shape[0]
    cdef size_t i
    nnl = l0
    if dl < -0.5 and nnl > 0:
        nnl -= 1
    elif dl > 0.5 and nnl < lmax - 1:
        nnl += 1
    nnp = p0
    if dp < -0.5 and nnp > 0:
        nnp -= 1
    elif dp > 0.5 and nnp < pmax - 1:
        nnp += 1
    for i in range(z_size):
        res[i] = data[i, nnl, nnp]


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void bil(const DTYPE_t[:, :, :] data, int l0, int p0, double dl, double dp, int lmax, int pmax, DTYPE_t[:] res) nogil:
    cdef int l_a, l_b, p_a, p_b
    cdef double w_l, w_p
    cdef size_t z_size = res.shape[0]
    cdef size_t i
    if dl < 0:
        l_a = max(0, l0 - 1)
        l_b = l0
        w_l = 1 + dl
    else:
        l_a = l0
        l_b = min(l0 + 1, lmax - 1)
        w_l = dl
    if dp < 0:
        p_a = max(0, p0 - 1)
        p_b = p0
        w_p = 1 + dp
    else:
        p_a = p0
        p_b = min(p0 + 1, pmax - 1)
        w_p = dp
    for i in range(z_size):
        res[i] = ((1 - w_l) * (1 - w_p) * data[i, l_a, p_a] +
                  (1 - w_l) * w_p * data[i, l_a, p_b] +
                  w_l * (1 - w_p) * data[i, l_b, p_a] +
                  w_l * w_p * data[i, l_b, p_b])

ctypedef void(*FN)(const DTYPE_t[:, :, :] data, int l0, int p0, double dl, double dp, int lmax, int pmax, DTYPE_t[:] res) nogil


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef one_step_gradient_search(np.ndarray[DTYPE_t, ndim=3] data,
                               np.ndarray[DTYPE_t, ndim=2] src_x,
                               np.ndarray[DTYPE_t, ndim=2] src_y,
                               np.ndarray[DTYPE_t, ndim=2] xl,
                               np.ndarray[DTYPE_t, ndim=2] xp,
                               np.ndarray[DTYPE_t, ndim=2] yl,
                               np.ndarray[DTYPE_t, ndim=2] yp,
                               np.ndarray[DTYPE_t, ndim=2] dst_x,
                               np.ndarray[DTYPE_t, ndim=2] dst_y,
                               method='bilinear'):
    """Gradient search, simple case variant."""
    cdef FN fun
    if method == 'bilinear':
        fun = bil
    else:
        fun = nn

    # change the output size (x_size, y_size) to match area_def.shape:
    # (lines,pixels)
    cdef size_t z_size = data.shape[0]
    cdef size_t y_size = dst_y.shape[0]
    cdef size_t x_size = dst_x.shape[1]

    # output image array --> needs to be (lines, pixels) --> y,x
    cdef np.ndarray[DTYPE_t, ndim= 3] image = np.full([z_size, y_size, x_size], np.nan, dtype=DTYPE)
    cdef np.ndarray[size_t, ndim= 1] elements = np.arange(x_size, dtype=np.uintp)

    one_step_gradient_search_no_gil(data,
                                    src_x, src_y,
                                    xl, xp, yl, yp,
                                    dst_x, dst_y,
                                    x_size, y_size,
                                    fun, image,
                                    elements)
    # return the output image
    return image


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void one_step_gradient_search_no_gil(const DTYPE_t[:, :, :] data,
                                          const DTYPE_t[:, :] src_x,
                                          const DTYPE_t[:, :] src_y,
                                          const DTYPE_t[:, :] xl,
                                          const DTYPE_t[:, :] xp,
                                          const DTYPE_t[:, :] yl,
                                          const DTYPE_t[:, :] yp,
                                          const DTYPE_t[:, :] dst_x,
                                          const DTYPE_t[:, :] dst_y,
                                          const size_t x_size,
                                          const size_t y_size,
                                          FN fun,
                                          DTYPE_t[:, :, :] image,
                                          size_t[:] elements) nogil:

    # pixel max ---> data is expected in [lines, pixels]
    cdef int pmax = data.shape[2] - 1
    cdef int lmax = data.shape[1] - 1
    # centre of input image - starting point
    cdef int p0 = pmax // 2
    cdef int l0 = lmax // 2
    cdef int last_p0 = p0
    cdef int last_l0 = l0
    # cdef int l0 = 0
    # intermediate variables:
    cdef int l_a, l_b, p_a, p_b
    cdef size_t i, j, elt
    cdef double dx, dy, d, dl, dp
    # number of iterations
    cdef int cnt = 0
    for i in range(y_size):
        # lines.reverse() --> swapped to elements - provide a reverse view of
        # the array
        elements = elements[::-1]
        for elt in range(x_size):
            j = elements[elt]
            if isinf(dst_x[i, j]):
                continue
            cnt = 0
            while True:
                cnt += 1
                # algorithm does not converge.
                if cnt > 5:
                    p0 = last_p0
                    l0 = last_l0
                    break
                # check we are within the input image bounds
                if l0 < lmax and l0 >= 0 and p0 < pmax and p0 >= 0:
                    # step size
                    dx = dst_x[i, j] - src_x[l0, p0]
                    dy = dst_y[i, j] - src_y[l0, p0]
                else:
                    # reset such that we are back in the input image bounds
                    if l0 >= lmax or l0 < 0 or p0 >= pmax or p0 < 0:
                        l0 = max(0, min(lmax - 1, l0))
                        p0 = max(0, min(pmax - 1, p0))
                        continue

                # distance from pixel/line to output location
                d = yl[l0, p0] * xp[l0, p0] - yp[l0, p0] * xl[l0, p0]
                if d == 0.0:
                    # There's no gradient, try again
                    continue
                dl = (xp[l0, p0] * dy - yp[l0, p0] * dx) / d
                dp = (yl[l0, p0] * dx - xl[l0, p0] * dy) / d
                # check that our distance to an output location is less than 1
                # pixel/line
                if fabs(dp) < 1 and fabs(dl) < 1:
                    last_p0 = p0
                    last_l0 = l0
                    #image[:, i, j] = fun(data, l0, p0, dl, dp, lmax, pmax)
                    fun(data, l0, p0, dl, dp, lmax, pmax, image[:, i, j])
                    # found our solution, next
                    break
                else:
                    # increment...
                    l0 = int(l0 + dl)
                    p0 = int(p0 + dp)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef fast_gradient_search_pg(np.ndarray[DTYPE_t, ndim=2] data,
                              np.ndarray[DTYPE_t, ndim=2] src_x,
                              np.ndarray[DTYPE_t, ndim=2] src_y,
                              np.ndarray[DTYPE_t, ndim=2] xl,
                              np.ndarray[DTYPE_t, ndim=2] xp,
                              np.ndarray[DTYPE_t, ndim=2] yl,
                              np.ndarray[DTYPE_t, ndim=2] yp,
                              np.ndarray[DTYPE_t, ndim=1] dst_x,
                              np.ndarray[DTYPE_t, ndim=1] dst_y,
                              method='bilinear'):
    """Gradient search, simple case variant."""
    #method = 'bilinear'
    cdef FN fun
    if method == 'bilinear':
        fun = bil
    else:
        fun = nn
    #print('method %s' % method)
    first_val = np.min((np.nanargmin(src_x), np.nanargmax(src_x),
                        np.nanargmin(src_y), np.nanargmax(src_y)))
    line, col = np.unravel_index(first_val, (src_x.shape[0], src_x.shape[1]))
    cdef size_t first_line, first_col
    first_line = line
    first_col = col
    #print(first_line, first_col)
    cdef size_t start_line = np.searchsorted(dst_y, src_y[first_line, first_col])
    cdef DTYPE_t[:, :] data_view = data
    cdef DTYPE_t[:, :] source_x_view = src_x
    cdef DTYPE_t[:, :] source_y_view = src_y
    cdef DTYPE_t[:, :] xl_view = xl
    cdef DTYPE_t[:, :] xp_view = xp
    cdef DTYPE_t[:, :] yl_view = yl
    cdef DTYPE_t[:, :] yp_view = yp
    # change the output size (x_size, y_size) to match area_def.shape:
    # (lines,pixels)
    cdef size_t y_size = len(dst_y)
    cdef size_t x_size = len(dst_x)
    cdef DTYPE_t[:] dst_x_view = dst_x
    cdef DTYPE_t[:] dst_y_view = dst_y

    # output image array --> needs to be (lines, pixels) --> y,x
    cdef np.ndarray[DTYPE_t, ndim= 2] image = np.full([y_size, x_size], np.nan, dtype=DTYPE)
    cdef DTYPE_t[:, :] image_view = image
    # this was a bit confusing -- "lines" was based on x_size; change this
    # variable to elements - make it a numpy array (long==numpy int dtype)
    cdef size_t[:] elements = np.arange(x_size, dtype=np.uintp)
    # Make a view to allow running without gil
    fast_gradient_search_ng(data_view,
                            source_x_view, source_y_view,
                            xl_view, xp_view, yl_view, yp_view,
                            x_size, y_size,
                            fun, image,
                            dst_x_view, dst_y_view,
                            elements, first_line,
                            first_col, start_line)
    # return the output image
    return image


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void fast_gradient_search_ng(DTYPE_t[:, :] data,
                                  DTYPE_t[:, :] src_x,
                                  DTYPE_t[:, :] src_y,
                                  DTYPE_t[:, :] xl,
                                  DTYPE_t[:, :] xp,
                                  DTYPE_t[:, :] yl,
                                  DTYPE_t[:, :] yp,
                                  size_t x_size, size_t y_size,
                                  FN fun,
                                  DTYPE_t[:, :] image,
                                  DTYPE_t[:] dst_x,
                                  DTYPE_t[:] dst_y,
                                  size_t[:] elements,
                                  size_t first_line,
                                  size_t first_col,
                                  size_t start_line) nogil:

    # pixel max ---> data is expected in [lines, pixels]
    cdef int pmax = data.shape[1] - 1
    cdef int lmax = data.shape[0] - 1
    # centre of input image - starting point
    cdef int p0 = pmax // 2
    cdef int l0 = lmax // 2
    # cdef int p0 = first_col
    # cdef int l0 = first_line
    cdef int last_p0 = p0
    cdef int last_l0 = l0
    # cdef int l0 = 0
    # intermediate variables:
    cdef int l_a, l_b, p_a, p_b
    cdef size_t i, j, elt
    cdef double dx, dy, d, dl, dp
    # number of iterations
    cdef int cnt = 0
    with nogil:
        for i in range(y_size):
            # lines.reverse() --> swapped to elements - provide a reverse view of
            # the array
            elements = elements[::-1]
            for elt in range(x_size):
                j = elements[elt]
                cnt = 0
                while True:
                    cnt += 1
                    # algorithm does not converge.
                    if cnt > 5:
                        p0 = last_p0
                        l0 = last_l0
                        break
                    # check we are within the input image bounds
                    if l0 < lmax and l0 >= 0 and p0 < pmax and p0 >= 0:
                        # while isnan(src_x[l0, p0]) and (dl > 1 or dp > 1):
                        #     dl /= 2
                        #     dp /= 2
                        #     l0 = int(l0 - dl)
                        #     p0 = int(p0 - dp)
                        # step size
                        dx = dst_x[j] - src_x[l0, p0]
                        dy = dst_y[i] - src_y[l0, p0]
                    else:
                        # while (l0 >= lmax or l0 < 0 or p0 >= pmax or p0 < 0 or isnan(src_x[l0, p0])) and (dl > 1 or dp > 1):
                        #     dl /= 2
                        #     dp /= 2
                        #     l0 = int(l0 - dl)
                        #     p0 = int(p0 - dp)
                        # reset such that we are back in the input image bounds
                        if l0 >= lmax or l0 < 0 or p0 >= pmax or p0 < 0:
                            l0 = max(0, min(lmax - 1, l0))
                            p0 = max(0, min(pmax - 1, p0))
                            continue
                        # if dl <= 1 and dp <= 1:
                        #     p0 = last_p0
                        #     l0 = last_l0
                        #     break
                            # continue
                        dx = dst_x[j] - src_x[l0, p0]
                        dy = dst_y[i] - src_y[l0, p0]

                    # distance from pixel/line to output location
                    d = yl[l0, p0] * xp[l0, p0] - yp[l0, p0] * xl[l0, p0]
                    if d == 0.0:
                        # There's no gradient, try again
                        continue
                    dl = (xp[l0, p0] * dy - yp[l0, p0] * dx) / d
                    dp = (yl[l0, p0] * dx - xl[l0, p0] * dy) / d
                    # check that our distance to an output location is less than 1
                    # pixel/line
                    if fabs(dp) < 1 and fabs(dl) < 1:
                        last_p0 = p0
                        last_l0 = l0
                        #image[y_size - 1 - i, j] = fun(data, l0, p0, dl, dp, lmax, pmax)
                        # found our solution, next
                        break
                    else:
                        # increment...
                        l0 = int(l0 + dl)
                        p0 = int(p0 + dp)


# old stuff

# @cython.boundscheck(False)
# @cython.wraparound(False)
# cpdef fast_gradient_search(np.ndarray[DTYPE_t, ndim=2] data, np.ndarray[DTYPE_t, ndim=2] src_x, np.ndarray[DTYPE_t, ndim=2] src_y, area_extent, size):
#       """Gradient search, simple case variant.
#       """
#       # area extent bounds --> area_def.area_extent
#       cdef double x_min, y_min, x_max, y_max
#       x_min, y_min, x_max, y_max = area_extent
#       # change the output size (x_size, y_size) to match area_def.shape:
#       # (lines,pixels)
#       cdef int x_size, y_size
#       y_size, x_size = size
#       # step in x direction (meters); x-output locations (centre of pixel);
#       # repeat for y
#       cdef double x_inc = (x_max - x_min) / x_size
#       cdef np.ndarray[DTYPE_t, ndim = 1] dst_x = np.arange(x_min + x_inc / 2, x_max, x_inc)
#       cdef double y_inc = (y_max - y_min) / y_size
#       cdef np.ndarray[DTYPE_t, ndim = 1] dst_y = np.arange(y_min + y_inc / 2, y_max, y_inc)
#       # output image array --> needs to be (lines, pixels) --> y,x
#       cdef np.ndarray[DTYPE_t, ndim = 2] image = np.zeros([y_size, x_size], dtype=DTYPE)
#       # gradient of output y/x grids (in pixel and line directions)
#       cdef np.ndarray[DTYPE_t, ndim = 2] yp, yl
#       yl, yp = np.gradient(src_y)
#       cdef np.ndarray[DTYPE_t, ndim = 2] xp, xl
#       xl, xp = np.gradient(src_x)
#       # pixel max ---> data is expected in [lines, pixels]
#       cdef int pmax = data.shape[1] - 1
#       cdef int lmax = data.shape[0] - 1
#       # centre of input image - starting point
#       cdef int p0 = pmax / 2
#       cdef int l0 = lmax / 2
#       # intermediate variables:
#       cdef int l_a, l_b, p_a, p_b, nnl, nnp
#       cdef size_t i, j, elt
#       cdef double dx, dy, d, dl, dp, w_l, w_p
#       # number of iterations
#       cdef int cnt = 0
#       # this was a bit confusing -- "lines" was based on x_size; change this
#       # variable to elements - make it a numpy array (long==numpy int dtype)
#       cdef np.ndarray[size_t, ndim = 1] npelements = np.arange(x_size, dtype=np.uintp)
#       # Make a view to allow running without gil
#       cdef size_t [:] elements = npelements
#       with nogil:
#         for i in range(y_size):
#             # lines.reverse() --> swapped to elements - provide a reverse view of
#             # the array
#             elements = elements[::-1]
#             for elt in range(elements.shape[0]):
#                 j = elements[elt]
#                 cnt = 0
#                 while True:
#                     cnt += 1
#                     # algorithm does not converge.
#                     if cnt > 5:
#                         break
#                     # check we are within the input image bounds
#                     if l0 < lmax and l0 >= 0 and p0 < pmax and p0 >= 0:
#                         # step size
#                         dx = dst_x[j] - src_x[l0, p0]
#                         dy = dst_y[i] - src_y[l0, p0]
#                     else:
#                         # reset such that we are back in the input image bounds
#                         l0 = max(0, min(lmax - 1, l0))
#                         p0 = max(0, min(pmax - 1, p0))
#                         continue
#                     # distance from pixel/line to output location
#                     d = yl[l0, p0] * xp[l0, p0] - yp[l0, p0] * xl[l0, p0]
#                     if d == 0.0:
#                         # There's no gradient, try again
#                         continue
#                     dl = (xp[l0, p0] * dy - yp[l0, p0] * dx) / d
#                     dp = (yl[l0, p0] * dx - xl[l0, p0] * dy) / d
#                     # check that our distance to an output location is less than 1
#                     # pixel/line
#                     if fabs(dp) < 1 and fabs(dl) < 1:
#                         # nearest neighbour
#                         # nnl = l0
#                         # if dl < -0.5 and nnl > 0:
#                         #     nnl -= 1
#                         # elif dl > 0.5 and nnl < lmax - 1:
#                         #     nnl += 1
#                         # nnp = p0
#                         # if dp < -0.5 and nnp > 0:
#                         #     nnp -= 1
#                         # elif dp > 0.5 and nnp < pmax - 1:
#                         #     nnp += 1
#                         # image[y_size - 1 - i, j] = data[nnl, nnp]
#                         # bilinear interpolation
#                         if dl < 0:
#                             l_a = max(0, l0 - 1)
#                             l_b = l0
#                             w_l = 1 + dl
#                         else:
#                             l_a = l0
#                             l_b = min(l0 + 1, lmax - 1)
#                             w_l = dl
#                         if dp < 0:
#                             p_a = max(0, p0 - 1)
#                             p_b = p0
#                             w_p = 1 + dp
#                         else:
#                             p_a = p0
#                             p_b = min(p0 + 1, pmax - 1)
#                             w_p = dp
#                         # assign image output... (was x_size in the first dimension
#                         # --> needs to be y_size)
#                         image[y_size - 1 - i, j] = ((1 - w_l) * (1 - w_p) * data[l_a, p_a] +
#                                                     (1 - w_l) * w_p * data[l_a, p_b] +
#                                                     w_l * (1 - w_p) * data[l_b, p_a] +
#                                                     w_l * w_p * data[l_b, p_b])
#                         # found our solution, next
#                         break
#                     else:
#                         # increment...
#                         l0 = int(l0 + dl)
#                         p0 = int(p0 + dp)
#       # return the output image
#       return image
#
#
#
#
# @cython.boundscheck(False)
# def fast_gradient_indices(np.ndarray[DTYPE_t, ndim=2] src_x, np.ndarray[DTYPE_t, ndim=2] src_y,
#                           np.ndarray[DTYPE_t, ndim=1] dst_x, np.ndarray[DTYPE_t, ndim=1] dst_y):
#     """Gradient search, simple case variant.
#     """
#     # area extent bounds --> area_def.area_extent
#     cdef double x_min, y_min, x_max, y_max
#     #x_min, y_min, x_max, y_max = area_extent
#     # change the output size (x_size, y_size) to match area_def.shape:
#     # (lines,pixels)
#     cdef int x_size, y_size
#     #y_size, x_size = size
#     x_size = dst_x.shape[0]
#     y_size = dst_y.shape[0]
#     # step in x direction (meters); x-output locations (centre of pixel);
#     # repeat for y
#     #cdef double x_inc = (x_max - x_min) / x_size
#     cdef double x_inc = (dst_x[-1] - dst_x[0]) / (x_size - 1)
#     #cdef np.ndarray[DTYPE_t, ndim = 1] dst_x = np.arange(x_min + x_inc / 2, x_max, x_inc)
#     #cdef double y_inc = (y_max - y_min) / y_size
#     cdef double y_inc = (dst_y[-1] - dst_y[0]) / (y_size - 1)
#     x_min = dst_x[0] - x_inc / 2
#     x_max = dst_x[-1] + x_inc / 2
#     y_min = dst_y[0] - y_inc / 2
#     y_max = dst_y[-1] + y_inc / 2
#
#     #cdef np.ndarray[DTYPE_t, ndim = 1] dst_y = np.arange(y_min + y_inc / 2, y_max, y_inc)
#     # output image array --> needs to be (lines, pixels) --> y,x
#     cdef np.ndarray[np.double_t, ndim = 3] indices = np.full([2, y_size, x_size], -2.0, dtype=np.double)
#     # gradient of output y/x grids (in pixel and line directions)
#     cdef np.ndarray[DTYPE_t, ndim = 2] yp, yl
#     yl, yp = np.gradient(src_y)
#     cdef np.ndarray[DTYPE_t, ndim = 2] xp, xl
#     xl, xp = np.gradient(src_x)
#     # pixel max ---> data is expected in [lines, pixels]
#     cdef int pmax = src_x.shape[1] - 1
#     cdef int lmax = src_x.shape[0] - 1
#     # centre of input image - starting point
#     cdef int p0 = pmax / 2
#     cdef int l0 = lmax / 2
#     # intermediate variables:
#     cdef int l_a, l_b, p_a, p_b, nnl, nnp
#     cdef size_t i, j
#     cdef double dx, dy, d, dl, dp, w_l, w_p
#     # number of iterations
#     cdef int cnt = 0
#     # this was a bit confusing -- "lines" was based on x_size; change this
#     # variable to elements - make it a numpy array (long==numpy int dtype)
#     cdef np.ndarray[size_t, ndim = 1] elements = np.arange(x_size, dtype=np.uintp)
#     for i in range(y_size):
#         # lines.reverse() --> swapped to elements - provide a reverse view of
#         # the array
#         elements = elements[::-1]
#         for j in elements:
#             cnt = 0
#             while True:
#                 cnt += 1
#                 # algorithm does not converge.
#                 if cnt > 5:
#                     break
#                 # check we are within the input image bounds
#                 if l0 < lmax and l0 >= 0 and p0 < pmax and p0 >= 0:
#                     # step size
#                     dx = dst_x[j] - src_x[l0, p0]
#                     dy = dst_y[i] - src_y[l0, p0]
#                 else:
#                     # reset such that we are back in the input image bounds
#                     l0 = max(0, min(lmax - 1, l0))
#                     p0 = max(0, min(pmax - 1, p0))
#                     continue
#                 # distance from pixel/line to output location
#                 d = yl[l0, p0] * xp[l0, p0] - yp[l0, p0] * xl[l0, p0]
#                 if d == 0.0:
#                     # There's no gradient, try again
#                     continue
#                 dl = (xp[l0, p0] * dy - yp[l0, p0] * dx) / d
#                 dp = (yl[l0, p0] * dx - xl[l0, p0] * dy) / d
#                 # check that our distance to an output location is less than 1
#                 # pixel/line
#                 if abs(dp) < 1 and abs(dl) < 1:
#                     indices[0, i, j] = max(min(l0 + dl, lmax - 1), 0)
#                     indices[1, i, j] = max(min(p0 + dp, pmax - 1), 0)
#                     # found our solution, next
#                     break
#                 else:
#                     # increment...
#                     l0 = int(l0 + dl)
#                     p0 = int(p0 + dp)
#     # return the output image
#     return indices
#
#
# @cython.boundscheck(False)
# def two_step_fast_gradient_search(np.ndarray[DTYPE_t, ndim=2] data, np.ndarray[DTYPE_t, ndim=2] src_x, np.ndarray[DTYPE_t, ndim=2] src_y, int chunk_size, area_extent, size):
#     """Gradient search, discontinuity handling variant (Modis)
#     """
#     # area extent bounds --> area_def.area_extent
#     cdef double x_min, y_min, x_max, y_max
#     x_min, y_min, x_max, y_max = area_extent
#     # change the output size (x_size, y_size) to match area_def.shape:
#     # (lines,pixels)
#     cdef int x_size, y_size
#     y_size, x_size = size
#     # step in x direction (meters); x-output locations (centre of pixel);
#     # repeat for y
#     cdef double x_inc = (x_max - x_min) / x_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_x = np.arange(x_min + x_inc / 2, x_max, x_inc)
#     cdef double y_inc = (y_max - y_min) / y_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_y = np.arange(y_min + y_inc / 2, y_max, y_inc)
#     # output image array --> needs to be (lines, pixels) --> y,x
#     cdef np.ndarray[DTYPE_t, ndim = 2] image = np.zeros([y_size, x_size], dtype=DTYPE)
#
#     cdef size_t i, j
#
#     # reduce the src_x and src_y arrays for inter-segment search
#     cdef np.ndarray[DTYPE_t, ndim = 2] reduced_x, reduced_y
#     # reduced_x = (src_x[4::chunk_size, :] + src_x[5::chunk_size]) / 2.0
#     # reduced_y = (src_y[4::chunk_size, :] + src_y[5::chunk_size]) / 2.0
#     reduced_x = (src_x[::chunk_size, :] + src_x[1::chunk_size] + src_x[2::chunk_size, :] + src_x[3::chunk_size] + src_x[4::chunk_size,
#                                                                                                                                        :] + src_x[5::chunk_size] + src_x[6::chunk_size, :] + src_x[7::chunk_size] + src_x[8::chunk_size, :] + src_x[9::chunk_size]) / chunk_size
#     reduced_y = (src_y[::chunk_size, :] + src_y[1::chunk_size] + src_y[2::chunk_size, :] + src_y[3::chunk_size] + src_y[4::chunk_size,
#                                                                                                                                        :] + src_y[5::chunk_size] + src_y[6::chunk_size, :] + src_y[7::chunk_size] + src_y[8::chunk_size, :] + src_y[9::chunk_size]) / chunk_size
#     cdef np.ndarray[DTYPE_t, ndim = 2] ryp, ryl
#     ryp, ryl = np.gradient(reduced_y)
#     cdef np.ndarray[DTYPE_t, ndim = 2] rxp, rxl
#     rxp, rxl = np.gradient(reduced_x)
#     rxp = -rxp
#     rxl = -rxl
#     ryp = -ryp
#     ryl = -ryl
#
#     # gradient of full output y/x grids (in pixel and line directions)
#     cdef np.ndarray[DTYPE_t, ndim = 2] fyp, fyl
#     cdef np.ndarray[DTYPE_t, ndim = 2] fxp, fxl
#     fyp = np.zeros_like(src_y)
#     fyl = np.zeros_like(src_y)
#     fxp = np.zeros_like(src_x)
#     fxl = np.zeros_like(src_x)
#     for i in range(data.shape[0] / chunk_size):
#         fyp[i * chunk_size:(i + 1) * chunk_size, :], fyl[i * chunk_size:(i + 1) *
#                                                          chunk_size, :] = np.gradient(src_y[i * chunk_size:(i + 1) * chunk_size, :])
#         fxp[i * chunk_size:(i + 1) * chunk_size, :], fxl[i * chunk_size:(i + 1) *
#                                                          chunk_size, :] = np.gradient(src_x[i * chunk_size:(i + 1) * chunk_size, :])
#
#     # fyp, fyl = np.gradient(src_y)
#     # fxp, fxl = np.gradient(src_x)
#     fxp = -fxp
#     fxl = -fxl
#     fyp = -fyp
#     fyl = -fyl
#
#     cdef np.ndarray[DTYPE_t, ndim = 2] yp, yl
#     cdef np.ndarray[DTYPE_t, ndim = 2] xp, xl
#     cdef np.ndarray[DTYPE_t, ndim = 2] array_x, array_y
#     # pixel max ---> data is expected in [lines, pixels]
#     cdef int pmax = data.shape[1] - 1
#     cdef int flmax = data.shape[0] - 1
#     cdef int rlmax = reduced_x.shape[0] - 1
#     cdef int lmax
#     # centre of input image - starting point
#     cdef int p0 = pmax / 2
#     cdef int l0 = flmax / 2
#     # intermediate variables:
#     cdef int l_a, l_b, p_a, p_b, nnl, nnp
#     cdef double dx, dy, d, dl, dp, w_l, w_p
#     # number of iterations
#     cdef int cnt = 0
#     cdef int adj = False
#     # this was a bit confusing -- "lines" was based on x_size; change this
#     # variable to elements - make it a numpy array (long==numpy int dtype)
#     cdef np.ndarray[size_t, ndim = 1] elements = np.arange(x_size, dtype=np.uintp)
#     lmax = rlmax
#     l0 /= chunk_size
#     xp, xl, yp, yl = rxp, rxl, ryp, ryl
#     array_x, array_y = reduced_x, reduced_y
#     for i in range(y_size):
#         # lines.reverse() --> swapped to elements - provide a reverse view of
#         # the array
#         elements = elements[::-1]
#         for j in elements:
#             cnt = 0
#             adj = False
#             while True:
#                 cnt += 1
#                 # algorithm does not converge, try jumping to the next chunk
#                 if cnt > 5:
#                     if not adj and l0 % chunk_size >= chunk_size / 2.0:
#                         l0 += chunk_size
#                         adj = True
#                         cnt = 0
#                     elif not adj and l0 % chunk_size < chunk_size / 2.0:
#                         l0 -= chunk_size
#                         adj = True
#                         cnt = 0
#                     else:
#                         break
#                 # check we are within the input image bounds
#                 if l0 < lmax and l0 >= 0 and p0 < pmax and p0 >= 0:
#                     # step size
#                     dx = dst_x[j] - array_x[l0, p0]
#                     dy = dst_y[i] - array_y[l0, p0]
#                 else:
#                     # reset such that we are back in the input image bounds
#                     l0 = max(0, min(lmax - 1, l0))
#                     p0 = max(0, min(pmax - 1, p0))
#                     break
#                 # distance from pixel/line to output location
#                 d = yl[l0, p0] * xp[l0, p0] - yp[l0, p0] * xl[l0, p0]
#                 dl = -(yl[l0, p0] * dx - xl[l0, p0] * dy) / d
#                 dp = (yp[l0, p0] * dx - xp[l0, p0] * dy) / d
#                 # check that our distance to an output location is less than 1
#                 # pixel/line
#                 if abs(dp) < 1 and abs(dl) < 1:
#                     if lmax == rlmax:  # switch to full resolution
#                         # nearest neighbour
#                         nnl = l0
#                         if dl < -0.5 and nnl > 0:
#                             nnl -= 1
#                         elif dl > 0.5 and nnl < lmax - 1:
#                             nnl += 1
#                         nnp = p0
#                         if dp < -0.5 and nnp > 0:
#                             nnp -= 1
#                         elif dp > 0.5 and nnp < pmax - 1:
#                             nnp += 1
#                         l0 = l0 * chunk_size + chunk_size / 2
#                         lmax = flmax
#                         xp, xl, yp, yl = fxp, fxl, fyp, fyl
#                         array_x, array_y = src_x, src_y
#                         continue
#                     # switch to next chunk if too close from bow-tie borders
#                     if not adj and l0 % chunk_size == chunk_size - 1:
#                         l0 += chunk_size / 2
#                         adj = True
#                         cnt = 0
#                         continue
#                     elif not adj and l0 % chunk_size == 0:
#                         l0 -= chunk_size / 2
#                         adj = True
#                         cnt = 0
#                         continue
#                     # bilinear interpolation
#                     if dl < 0:
#                         l_a = max(0, l0 - 1)
#                         l_b = l0
#                         w_l = 1 + dl
#                     else:
#                         l_a = l0
#                         l_b = min(l0 + 1, lmax - 1)
#                         w_l = dl
#                     if dp < 0:
#                         p_a = max(0, p0 - 1)
#                         p_b = p0
#                         w_p = 1 + dp
#                     else:
#                         p_a = p0
#                         p_b = min(p0 + 1, pmax - 1)
#                         w_p = dp
#                     # assign image output... (was x_size in the first dimension
#                     # --> needs to be y_size)
#                     image[y_size - 1 - i, j] = ((1 - w_l) * (1 - w_p) * data[l_a, p_a] +
#                                                 (1 - w_l) * w_p * data[l_a, p_b] +
#                                                 w_l * (1 - w_p) * data[l_b, p_a] +
#                                                 w_l * w_p * data[l_b, p_b])
#                     # image[y_size - 1 - i, j] = data[l0, p0]
#                     # found our solution, next
#                     break
#                 else:
#                     # increment...
#                     l0 = int(l0 + dl)
#                     p0 = int(p0 + dp)
#     # return the output image
#     return image
#
#
# @cython.boundscheck(False)
# def fast_gradient_search_with_mask(np.ndarray[DTYPE_t, ndim=2] data, np.ndarray[DTYPE_t, ndim=2] src_x, np.ndarray[DTYPE_t, ndim=2] src_y, area_extent, size, np.ndarray[np.uint8_t, ndim=2] mask):
#     """Gradient search, simple case variant.
#     """
#     # area extent bounds --> area_def.area_extent
#     cdef double x_min, y_min, x_max, y_max
#     x_min, y_min, x_max, y_max = area_extent
#     # change the output size (x_size, y_size) to match area_def.shape:
#     # (lines,pixels)
#     cdef int x_size, y_size
#     y_size, x_size = size
#     # step in x direction (meters); x-output locations (centre of pixel);
#     # repeat for y
#     cdef double x_inc = (x_max - x_min) / x_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_x = np.arange(x_min + x_inc / 2, x_max, x_inc)
#     cdef double y_inc = (y_max - y_min) / y_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_y = np.arange(y_min + y_inc / 2, y_max, y_inc)
#     # output image array --> needs to be (lines, pixels) --> y,x
#     cdef np.ndarray[DTYPE_t, ndim = 2] image = np.zeros([y_size, x_size], dtype=DTYPE)
#     # gradient of output y/x grids (in pixel and line directions)
#     cdef np.ndarray[DTYPE_t, ndim = 2] yp, yl
#     yl, yp = np.gradient(src_y)
#     cdef np.ndarray[DTYPE_t, ndim = 2] xp, xl
#     xl, xp = np.gradient(src_x)
#     # pixel max ---> data is expected in [lines, pixels]
#     cdef int pmax = data.shape[1] - 1
#     cdef int lmax = data.shape[0] - 1
#     # centre of input image - starting point
#     cdef int p0 = pmax / 2
#     cdef int l0 = lmax / 2
#     cdef int oldp, oldl
#     # intermediate variables:
#     cdef int l_a, l_b, p_a, p_b, nnl, nnp
#     cdef size_t i, j
#     cdef double dx, dy, d, dl, dp, w_l, w_p
#     # number of iterations
#     cdef int cnt = 0
#     cdef int inc = 0
#     # this was a bit confusing -- "lines" was based on x_size; change this
#     # variable to elements - make it a numpy array (long==numpy int dtype)
#     cdef np.ndarray[size_t, ndim = 1] elements = np.arange(x_size, dtype=np.uintp)
#     for i in range(y_size):
#         # lines.reverse() --> swapped to elements - provide a reverse view of
#         # the array
#         elements = elements[::-1]
#         for j in elements:
#             cnt = 0
#             while True:
#                 cnt += 1
#                 # algorithm does not converge.
#                 if cnt > 15:
#                     break
#                 # step size
#                 dx = dst_x[j] - src_x[l0, p0]
#                 dy = dst_y[i] - src_y[l0, p0]
#
#                 # distance from pixel/line to output location
#                 d = yl[l0, p0] * xp[l0, p0] - yp[l0, p0] * xl[l0, p0]
#                 dl = (xp[l0, p0] * dy - yp[l0, p0] * dx) / d
#                 dp = (yl[l0, p0] * dx - xl[l0, p0] * dy) / d
#                 # check that our distance to an output location is less than 1
#                 # pixel/line
#                 if abs(dp) < 1 and abs(dl) < 1:
#                     image[y_size - 1 - i, j] = data[l0, p0]
#                     # found our solution, next
#                     break
#                 else:
#                     # increment...
#                     l0 = int(l0 + dl)
#                     p0 = int(p0 + dp)
#                     # if dl > 0 and l0 % chunk_size == chunk_size - 1:
#                     #    l0 += 2
#                     # if dl < 0 and l0 % chunk_size == 0:
#                     #    l0 -= 2
#
#                     # oldp = p0
#                     # oldl = l0
#                     # p0 = max(0, min(pmax - 1, int(p0 + dp)))
#                     # l0 = int(l0 + dl)
#                     # inc = int(dl)
#                     # if inc > 0:
#                     #     while (inc > 0 or mask[l0, p0] != 0) and l0 < lmax:
#                     #         if mask[l0, p0] == 0:
#                     #             inc -= 1
#                     #         l0 += 1
#                     #     if l0 >= lmax:
#                     #         l0 = oldl
#                     # else:
#                     #     while (inc < 0 or mask[l0, p0] != 0) and l0 >= 0:
#                     #         if mask[l0, p0] == 0:
#                     #             inc += 1
#                     #         l0 -= 1
#                     #     if l0 < 0:
#                     #         l0 = oldl
#                     # if oldp == p0 and oldl == l0:
#                     #     break
#     # return the output image
#     return image
#
#
# @cython.boundscheck(False)
# def two_step_fast_gradient_search_with_mask(np.ndarray[DTYPE_t, ndim=2] data, np.ndarray[DTYPE_t, ndim=2] src_x, np.ndarray[DTYPE_t, ndim=2] src_y, int chunk_size, area_extent, size, np.ndarray[np.uint8_t, ndim=2] mask):
#     """Gradient search, discontinuity handling variant (Modis)
#     """
#     # area extent bounds --> area_def.area_extent
#     cdef double x_min, y_min, x_max, y_max
#     x_min, y_min, x_max, y_max = area_extent
#     # change the output size (x_size, y_size) to match area_def.shape:
#     # (lines,pixels)
#     cdef int x_size, y_size
#     y_size, x_size = size
#     # step in x direction (meters); x-output locations (centre of pixel);
#     # repeat for y
#     cdef double x_inc = (x_max - x_min) / x_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_x = np.arange(x_min + x_inc / 2, x_max, x_inc)
#     cdef double y_inc = (y_max - y_min) / y_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_y = np.arange(y_min + y_inc / 2, y_max, y_inc)
#     # output image array --> needs to be (lines, pixels) --> y,x
#     cdef np.ndarray[DTYPE_t, ndim = 2] image = np.zeros([y_size, x_size], dtype=DTYPE)
#
#     cdef size_t i, j
#
#     # reduce the src_x and src_y arrays for inter-segment search
#     cdef np.ndarray[DTYPE_t, ndim = 2] reduced_x, reduced_y
#     reduced_x = np.zeros([data.shape[0] / chunk_size, data.shape[1]],
#                          dtype=DTYPE)
#     reduced_y = np.zeros([data.shape[0] / chunk_size, data.shape[1]],
#                          dtype=DTYPE)
#     for i in range(chunk_size):
#         reduced_x += src_x[i::chunk_size, :]
#         reduced_y += src_y[i::chunk_size, :]
#     reduced_x /= chunk_size
#     reduced_y /= chunk_size
#     cdef np.ndarray[DTYPE_t, ndim = 2] ryp, ryl
#     ryp, ryl = np.gradient(reduced_y)
#     cdef np.ndarray[DTYPE_t, ndim = 2] rxp, rxl
#     rxp, rxl = np.gradient(reduced_x)
#     rxp = -rxp
#     rxl = -rxl
#     ryp = -ryp
#     ryl = -ryl
#
#     # gradient of full output y/x grids (in pixel and line directions)
#     cdef np.ndarray[DTYPE_t, ndim = 2] fyp, fyl
#     cdef np.ndarray[DTYPE_t, ndim = 2] fxp, fxl
#     fyp = np.zeros_like(src_y)
#     fyl = np.zeros_like(src_y)
#     fxp = np.zeros_like(src_x)
#     fxl = np.zeros_like(src_x)
#     for i in range(data.shape[0] / chunk_size):
#         fyp[i * chunk_size:(i + 1) * chunk_size, :], fyl[i * chunk_size:(i + 1) *
#                                                          chunk_size, :] = np.gradient(src_y[i * chunk_size:(i + 1) * chunk_size, :])
#         fxp[i * chunk_size:(i + 1) * chunk_size, :], fxl[i * chunk_size:(i + 1) *
#                                                          chunk_size, :] = np.gradient(src_x[i * chunk_size:(i + 1) * chunk_size, :])
#
#     # fyp, fyl = np.gradient(src_y)
#     # fxp, fxl = np.gradient(src_x)
#     fxp = -fxp
#     fxl = -fxl
#     fyp = -fyp
#     fyl = -fyl
#
#     cdef np.ndarray[DTYPE_t, ndim = 2] yp, yl
#     cdef np.ndarray[DTYPE_t, ndim = 2] xp, xl
#     cdef np.ndarray[DTYPE_t, ndim = 2] array_x, array_y
#     # pixel max ---> data is expected in [lines, pixels]
#     cdef int pmax = data.shape[1] - 1
#     cdef int flmax = data.shape[0] - 1
#     cdef int rlmax = reduced_x.shape[0] - 1
#     cdef int lmax
#     # centre of input image - starting point
#     cdef int p0 = pmax / 2
#     cdef int l0 = flmax / 2
#     # intermediate variables:
#     cdef int nnl, nnp
#     cdef int l_a, l_b, l_c, l_d, p_a, p_b, p_c, p_d
#     cdef double dx, dy, d, dl, dp, w_l, w_p
#     # number of iterations
#     cdef int cnt = 0
#     cdef int adj = False
#     cdef int undefined = 0
#     cdef int sdp = 1
#     cdef int sdl = 1
#     cdef int n_l, n_p
#     cdef double distance
#     cdef int masked = False
#     # this was a bit confusing -- "lines" was based on x_size; change this
#     # variable to elements - make it a numpy array (long==numpy int dtype)
#     cdef np.ndarray[size_t, ndim = 1] elements = np.arange(x_size, dtype=np.uintp)
#     lmax = rlmax
#     l0 /= chunk_size
#     xp, xl, yp, yl = rxp, rxl, ryp, ryl
#     array_x, array_y = reduced_x, reduced_y
#
#     for i in range(y_size):
#         # lines.reverse() --> swapped to elements - provide a reverse view of
#         # the array
#         elements = elements[::-1]
#         for j in elements:
#             cnt = 0
#             adj = False
#             while True:
#                 cnt += 1
#                 # algorithm does not converge
#                 if cnt > 5:
#                     break
#
#                 # check we are within the input image bounds
#                 if not (l0 < lmax and l0 >= 0 and p0 < pmax and p0 >= 0):
#                     # reset such that we are back in the input image bounds
#                     l0 = max(0, min(lmax - 1, l0))
#                     p0 = max(0, min(pmax - 1, p0))
#                 # step size
#                 dx = dst_x[j] - array_x[l0, p0]
#                 dy = dst_y[i] - array_y[l0, p0]
#
#                 # distance from pixel/line to output location
#                 d = yl[l0, p0] * xp[l0, p0] - yp[l0, p0] * xl[l0, p0]
#                 dl = -(yl[l0, p0] * dx - xl[l0, p0] * dy) / d
#                 dp = (yp[l0, p0] * dx - xp[l0, p0] * dy) / d
#
#                 # check that our distance to an output location is less than 1
#                 # pixel/line
#                 if ((abs(dp) < 1) and (abs(dl) < 1) and
#                         mask[l0, p0] == 0):
#                     # not (mask[l0, p0] != 0 or
#                     #     (l0 < lmax - 1 and mask[l0 + 1, p0] != 0) or
#                     #     (l0 > 0 and mask[l0 - 1, p0] != 0))):
#                     if lmax == rlmax:  # switch to full resolution
#                         print "switching to full res", i, j
#                         # nearest neighbour
#                         l0 = l0 * chunk_size + chunk_size / 2
#                         lmax = flmax
#                         xp, xl, yp, yl = fxp, fxl, fyp, fyl
#                         array_x, array_y = src_x, src_y
#                         cnt = 0
#                         continue
#
#                     # crude
#                     #image[y_size - 1 - i, j] = data[l0, p0]
#
#                     # nearest neighbour
#                     n_l = max(0, min(lmax - 1, int(l0 + dl)))
#                     n_p = max(0, min(pmax - 1, int(p0 + dp)))
#
#                     if mask[n_l, n_p] != 0:
#                         image[y_size - 1 - i, j] = data[l0, p0]
#                     else:
#                         image[y_size - 1 - i, j] = data[n_l, n_p]
#
#                     # bilinear interpolation
#
#                     # if dl < 0:
#                     #     sdl = -1
#                     # else:
#                     #     sdl = 1
#                     # if dp < 0:
#                     #     sdp = -1
#                     # else:
#                     #     sdp = 1
#
#                     # l_a, p_a = l0, p0
#                     # masked = False
#
#                     # l_b, p_b = l0, max(0, min(p0 + sdp, pmax - 1))
#                     # while (l_b >= 0 and l_b < lmax
#                     #        and mask[l_b, p_b] != 0):
#                     #     l_b -= sdl
#                     #     masked = True
#                     # l_b = max(0, min(l_b, lmax - 1))
#                     # if mask[l_b, p_b] != 0:
#                     #     l_b, p_b = l_a, p_a
#
#                     # l_c, p_c = (max(0, min(l0 + sdl, lmax - 1)),
#                     #             max(0, min(p0 + sdp, pmax - 1)))
#                     # while (l_c >= 0 and l_c < lmax and
#                     #        mask[l_c, p_c] != 0):
#                     #     l_c += sdl
#                     #     masked = True
#                     # l_c = max(0, min(l_c, lmax - 1))
#                     # if mask[l_c, p_c] != 0:
#                     #     l_c, p_c = l_a, p_a
#
#                     # l_d, p_d = max(0, min(l0 + sdl, lmax - 1)), p0
#                     # while (l_d >= 0 and l_d < lmax and
#                     #        mask[l_d, p_d] != 0):
#                     #     l_d += sdl
#                     #     masked = True
#                     # l_d = max(0, min(l_d, lmax - 1))
#                     # if mask[l_d, p_d] != 0:
#                     #     l_d, p_d = l_a, p_a
#
#                     # if masked and abs(l_a - l_d) > 9:
#                     #     print "masked: lines", l_a, l_b, l_c, l_d,
#                     #     print "cols:", p_a, p_b, p_c, p_d
#
#                     # recompute dl and dp
#                     # if masked and l_a != l_d and p_a != p_b:
#                     #     print "before", dl, dp
#                     #     d = ((src_y[l_a, p_a] - src_y[l_d, p_d]) *
#                     #          (src_x[l_a, p_a] - src_x[l_b, p_b]) -
#                     #          (src_y[l_a, p_a] - src_y[l_b, p_b]) *
#                     #          (src_x[l_a, p_a] - src_x[l_d, p_d]))
#                     #     dp = -((src_y[l_a, p_a] - src_y[l_d, p_d]) *
#                     #            dx - (src_x[l_a, p_a] - src_x[l_d, p_d]) * dy) / d
#                     #     dl = ((src_y[l_a, p_a] - src_y[l_b, p_b]) *
#                     #           dx - (src_x[l_a, p_a] - src_x[l_b, p_b]) * dy) / d
#                     #     print "after", dl, dp
#                     #     dp = min(1, max(dp, -1))
#                     #     dl = min(1, max(dl, -1))
#
#                     # w_l = 1 - abs(dl)
#                     # w_p = 1 - abs(dp)
#
#                     # image[y_size - 1 - i, j] = ((1 - w_l) * (1 - w_p) * data[l_c, p_c] +
#                     #                             (1 - w_l) * w_p * data[l_d, p_d] +
#                     #                             w_l * (1 - w_p) * data[l_b, p_b] +
#                     #                             w_l * w_p * data[l_a, p_a])
#
#                     # we found our solution, next
#                     break
#                 else:
#                     # increment...
#                     l0 = int(l0 + dl)
#                     p0 = int(p0 + dp)
#                     # if dl > 0 and l0 % chunk_size >= chunk_size - 1:
#                     #    l0 += 2
#                     # if dl < 0 and l0 % chunk_size < 1:
#                     #    l0 -= 2
#                     if l0 >= lmax or l0 < 0 or p0 >= pmax or p0 < 0:
#                         continue
#
#                     # if dl > 0 and (mask[l0, p0] != 0 or
#                     #                (l0 < lmax - 1 and
#                     #                 mask[l0 + 1, p0] != 0)):
#                     #     while (l0 < lmax - 1 and
#                     #            (mask[l0, p0] != 0 or
#                     #             mask[l0 + 1, p0] != 0)):
#                     #         l0 += 1
#                     #     l0 = min(lmax - 1, l0)
#
#                     # if dl < 0 and (mask[l0, p0] != 0 or
#                     #                (l0 > 0 and
#                     #                 mask[l0 - 1, p0] != 0)):
#
#                     #     while (l0 > 0 and
#                     #            (mask[l0, p0] != 0 or
#                     #             mask[l0 - 1, p0] != 0)):
#                     #         l0 -= 1
#                     #     l0 = max(0, l0)
#                     if dl > 0 and mask[l0, p0] != 0:
#                         while (l0 < lmax - 1 and
#                                mask[l0, p0] != 0):
#                             l0 += 1
#                         l0 = min(lmax - 1, l0)
#
#                     if dl < 0 and mask[l0, p0] != 0:
#
#                         while (l0 > 0 and
#                                mask[l0, p0] != 0):
#                             l0 -= 1
#                         l0 = max(0, l0)
#
#     # return the output image
#     print "undefined", undefined
#     return image
#
#
# #############################
# ########################
# ######################
# #############
# #
# #  Old stuff
# #
# ######
# ##
#
# @cython.boundscheck(False)
# def two_step_fast_gradient_search_with_mask_old(np.ndarray[DTYPE_t, ndim=2] data, np.ndarray[DTYPE_t, ndim=2] src_x, np.ndarray[DTYPE_t, ndim=2] src_y, int chunk_size, area_extent, size, np.ndarray[np.uint8_t, ndim=2] mask):
#     """Gradient search, discontinuity handling variant (Modis)
#     """
#     # area extent bounds --> area_def.area_extent
#     cdef double x_min, y_min, x_max, y_max
#     x_min, y_min, x_max, y_max = area_extent
#     # change the output size (x_size, y_size) to match area_def.shape:
#     # (lines,pixels)
#     cdef int x_size, y_size
#     y_size, x_size = size
#     # step in x direction (meters); x-output locations (centre of pixel);
#     # repeat for y
#     cdef double x_inc = (x_max - x_min) / x_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_x = np.arange(x_min + x_inc / 2, x_max, x_inc)
#     cdef double y_inc = (y_max - y_min) / y_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_y = np.arange(y_min + y_inc / 2, y_max, y_inc)
#     # output image array --> needs to be (lines, pixels) --> y,x
#     cdef np.ndarray[DTYPE_t, ndim = 2] image = np.zeros([y_size, x_size], dtype=DTYPE)
#
#     cdef size_t i, j
#
#     # reduce the src_x and src_y arrays for inter-segment search
#     cdef np.ndarray[DTYPE_t, ndim = 2] reduced_x, reduced_y
#     # reduced_x = (src_x[4::chunk_size, :] + src_x[5::chunk_size]) / 2.0
#     # reduced_y = (src_y[4::chunk_size, :] + src_y[5::chunk_size]) / 2.0
#     # reduced_x = (src_x[::chunk_size, :] + src_x[1::chunk_size] + src_x[2::chunk_size, :] + src_x[3::chunk_size] + src_x[4::chunk_size,:] + src_x[5::chunk_size] + src_x[6::chunk_size, :] + src_x[7::chunk_size] + src_x[8::chunk_size, :] + src_x[9::chunk_size]) / chunk_size
#     # reduced_y = (src_y[::chunk_size, :] + src_y[1::chunk_size] +
#     # src_y[2::chunk_size, :] + src_y[3::chunk_size] +
#     # src_y[4::chunk_size, :] + src_y[5::chunk_size] +
#     # src_y[6::chunk_size, :] + src_y[7::chunk_size] +
#     # src_y[8::chunk_size, :] + src_y[9::chunk_size]) / chunk_size
#     reduced_x = np.zeros([data.shape[0] / chunk_size, data.shape[1]],
#                          dtype=DTYPE)
#     reduced_y = np.zeros([data.shape[0] / chunk_size, data.shape[1]],
#                          dtype=DTYPE)
#     for i in range(chunk_size):
#         reduced_x += src_x[i::chunk_size, :]
#         reduced_y += src_y[i::chunk_size, :]
#     reduced_x /= chunk_size
#     reduced_y /= chunk_size
#     cdef np.ndarray[DTYPE_t, ndim = 2] ryp, ryl
#     ryp, ryl = np.gradient(reduced_y)
#     cdef np.ndarray[DTYPE_t, ndim = 2] rxp, rxl
#     rxp, rxl = np.gradient(reduced_x)
#     rxp = -rxp
#     rxl = -rxl
#     ryp = -ryp
#     ryl = -ryl
#
#     # gradient of full output y/x grids (in pixel and line directions)
#     cdef np.ndarray[DTYPE_t, ndim = 2] fyp, fyl
#     cdef np.ndarray[DTYPE_t, ndim = 2] fxp, fxl
#     fyp = np.zeros_like(src_y)
#     fyl = np.zeros_like(src_y)
#     fxp = np.zeros_like(src_x)
#     fxl = np.zeros_like(src_x)
#     for i in range(data.shape[0] / chunk_size):
#         fyp[i * chunk_size:(i + 1) * chunk_size, :], fyl[i * chunk_size:(i + 1) *
#                                                          chunk_size, :] = np.gradient(src_y[i * chunk_size:(i + 1) * chunk_size, :])
#         fxp[i * chunk_size:(i + 1) * chunk_size, :], fxl[i * chunk_size:(i + 1) *
#                                                          chunk_size, :] = np.gradient(src_x[i * chunk_size:(i + 1) * chunk_size, :])
#
#     # fyp, fyl = np.gradient(src_y)
#     # fxp, fxl = np.gradient(src_x)
#     fxp = -fxp
#     fxl = -fxl
#     fyp = -fyp
#     fyl = -fyl
#
#     cdef np.ndarray[DTYPE_t, ndim = 2] yp, yl
#     cdef np.ndarray[DTYPE_t, ndim = 2] xp, xl
#     cdef np.ndarray[DTYPE_t, ndim = 2] array_x, array_y
#     # pixel max ---> data is expected in [lines, pixels]
#     cdef int pmax = data.shape[1] - 1
#     cdef int flmax = data.shape[0] - 1
#     cdef int rlmax = reduced_x.shape[0] - 1
#     cdef int lmax
#     # centre of input image - starting point
#     cdef int p0 = pmax / 2
#     cdef int l0 = flmax / 2
#     # intermediate variables:
#     cdef int nnl, nnp
#     cdef int l_a00, l_a01, l_a10, l_a11, p_a00, p_a01, p_a10, p_a11
#     cdef double dx, dy, d, dl, dp, w_l, w_p
#     # number of iterations
#     cdef int cnt = 0
#     cdef int adj = False
#     cdef int undefined = 0
#     cdef double distance
#     # this was a bit confusing -- "lines" was based on x_size; change this
#     # variable to elements - make it a numpy array (long==numpy int dtype)
#     cdef np.ndarray[size_t, ndim = 1] elements = np.arange(x_size, dtype=np.uintp)
#     lmax = rlmax
#     l0 /= chunk_size
#     xp, xl, yp, yl = rxp, rxl, ryp, ryl
#     array_x, array_y = reduced_x, reduced_y
#     for i in range(y_size):
#         # lines.reverse() --> swapped to elements - provide a reverse view of
#         # the array
#         elements = elements[::-1]
#         for j in elements:
#             cnt = 0
#             adj = False
#             while True:
#                 cnt += 1
#                 # algorithm does not converge, try jumping to the next chunk
#                 if cnt > 5:
#                     break
#                     # if adj and l0 % chunk_size >= chunk_size / 2.0:
#                     #     l0 += chunk_size
#                     #     adj = True
#                     #     cnt = 0
#                     # elif not adj and l0 % chunk_size < chunk_size / 2.0:
#                     #     l0 -= chunk_size
#                     #     adj = True
#                     #     cnt = 0
#                     # else:
#                     # distance = np.sqrt((src_y[l0, p0] - dst_y[i]) ** 2 +
#                     # (src_x[l0, p0] - dst_x[j]) ** 2)
#                     # TODO: this should be done dynamically or from arg
#                     # if distance < 1:
#                     # image[y_size - 1 - i, j] = data[l0, p0]
#                     # else:
#                     # undefined += 1
#                     #     break
#
#                 # check we are within the input image bounds
#                 if l0 < lmax and l0 >= 0 and p0 < pmax and p0 >= 0:
#                     if mask[l0, p0] != 0:
#                         if dl > 0:
#                             while l0 < lmax and mask[l0, p0] != 0:
#                                 l0 += 1
#                             if l0 >= lmax:
#                                 l0 = lmax - 1
#                                 break
#                         else:
#                             while l0 >= 0 and mask[l0, p0] != 0:
#                                 l0 -= 1
#                             if l0 < 0:
#                                 l0 = 0
#                                 break
#
#                     # step size
#                     dx = dst_x[j] - array_x[l0, p0]
#                     dy = dst_y[i] - array_y[l0, p0]
#                 else:
#                     # reset such that we are back in the input image bounds
#                     l0 = max(0, min(lmax - 1, l0))
#                     p0 = max(0, min(pmax - 1, p0))
#                     dx = dst_x[j] - array_x[l0, p0]
#                     dy = dst_y[i] - array_y[l0, p0]
#
#                 # distance from pixel/line to output location
#                 d = yl[l0, p0] * xp[l0, p0] - yp[l0, p0] * xl[l0, p0]
#                 dl = -(yl[l0, p0] * dx - xl[l0, p0] * dy) / d
#                 dp = (yp[l0, p0] * dx - xp[l0, p0] * dy) / d
#
#                 # take care of pixels outside
#                 if ((l0 == lmax - 1 and dl > 0.5) or
#                         (l0 == 0 and dl < -0.5) or
#                         (p0 == pmax - 1 and dp > 0.5) or
#                         (p0 == 0 and dp < -0.5)):
#                     break
#
#                 # check that our distance to an output location is less than 1
#                 # pixel/line
#                 if (abs(dp) < 1) and (abs(dl) < 1):
#                     if lmax == rlmax:  # switch to full resolution
#                         print "switching to full res", i, j
#                         # nearest neighbour
#                         l0 = l0 * chunk_size + chunk_size / 2
#                         lmax = flmax
#                         xp, xl, yp, yl = fxp, fxl, fyp, fyl
#                         array_x, array_y = src_x, src_y
#                         cnt = 0
#                         continue
#                     # switch to next chunk if too close from bow-tie borders
#                     # if not adj and l0 % chunk_size == chunk_size - 1:
#                     #     l0 += chunk_size / 2
#                     #     adj = True
#                     #     cnt = 0
#                     #     continue
#                     # elif not adj and l0 % chunk_size == 0:
#                     #     l0 -= chunk_size / 2
#                     #     adj = True
#                     #     cnt = 0
#                     #     continue
#                     # check if the data is masked
#                     if mask[l0, p0] != 0:
#                         # if dl > 0:
#                         #     while mask[l0, p0] != 0:
#                         #         l0 += 1
#                         # else:
#                         #     while mask[l0, p0] != 0:
#                         #         l0 -= 1
#                         if l0 % chunk_size > chunk_size / 2:
#                             while l0 < lmax and mask[l0, p0] != 0:
#                                 l0 += 1
#                             if l0 >= lmax:
#                                 l0 = lmax - 1
#                                 break
#                         else:
#                             while l0 >= 0 and mask[l0, p0] != 0:
#                                 l0 -= 1
#                             if l0 < 0:
#                                 l0 = 0
#                                 break
#                         cnt -= 1
#                         continue
#
#                     # image[y_size - 1 - i, j] = data[l0, p0]
#
#                     # bilinear interpolation
#                     # if dl < 0:
#                     #     l_a = max(0, l0 - 1)
#                     #     l_b = l0
#                     #     w_l = 1 + dl
#                     # else:
#                     #     l_a = l0
#                     #     l_b = min(l0 + 1, lmax - 1)
#                     #     w_l = dl
#                     # if dp < 0:
#                     #     p_a = max(0, p0 - 1)
#                     #     p_b = p0
#                     #     w_p = 1 + dp
#                     # else:
#                     #     p_a = p0
#                     #     p_b = min(p0 + 1, pmax - 1)
#                     #     w_p = dp
#
#                     if dl < 0 and dp < 0:
#                         w_l = 1 + dl
#                         w_p = 1 + dp
#                         l_a00 = l0 - 1
#                         p_a00 = p0 - 1
#                         l_a01 = l0 - 1
#                         p_a01 = p0
#                         l_a10 = l0
#                         p_a10 = p0 - 1
#                         l_a11 = l0
#                         p_a11 = p0
#                         while l_a01 > 0 and mask[l_a01, p_a01] != 0:
#                             l_a01 -= 1
#                         if mask[l_a00, p_a00] != 0:
#                             l_a00 = l_a01
#                             p_a00 += 1
#                         if mask[l_a10, p_a10] != 0:
#                             p_a10 += 1
#
#                     if dl > 0 and dp < 0:
#                         w_l = dl
#                         w_p = 1 + dp
#                         l_a00 = l0
#                         p_a00 = p0 - 1
#                         l_a01 = l0
#                         p_a01 = p0
#                         l_a10 = l0 + 1
#                         p_a10 = p0 - 1
#                         l_a11 = l0 + 1
#                         p_a11 = p0
#                         if mask[l_a00, p_a00] != 0:
#                             p_a00 += 1
#                         while l_a11 < lmax - 1 and mask[l_a11, p_a11] != 0:
#                             l_a11 += 1
#                         if mask[l_a10, p_a10] != 0:
#                             l_a10 = l_a11
#                             p_a10 += 1
#
#                     if dl > 0 and dp > 0:
#                         w_l = dl
#                         w_p = dp
#                         l_a00 = l0
#                         p_a00 = p0
#                         l_a01 = l0
#                         p_a01 = p0 + 1
#                         l_a10 = l0 + 1
#                         p_a10 = p0
#                         l_a11 = l0 + 1
#                         p_a11 = p0 + 1
#                         if mask[l_a01, p_a01] != 0:
#                             p_a01 -= 1
#                         while l_a10 < lmax - 1 and mask[l_a10, p_a10] != 0:
#                             l_a10 += 1
#                         if mask[l_a11, p_a11] != 0:
#                             l_a11 = l_a10
#                             p_a11 -= 1
#
#                     if dl < 0 and dp > 0:
#                         w_l = 1 + dl
#                         w_p = dp
#                         l_a00 = l0 - 1
#                         p_a00 = p0
#                         l_a01 = l0 - 1
#                         p_a01 = p0 + 1
#                         l_a10 = l0
#                         p_a10 = p0
#                         l_a11 = l0
#                         p_a11 = p0 + 1
#
#                         while l_a00 > 0 and mask[l_a00, p_a00] != 0:
#                             l_a00 -= 1
#                         if mask[l_a01, p_a01] != 0:
#                             l_a01 = l_a00
#                             p_a01 -= 1
#                         if mask[l_a11, p_a11] != 0:
#                             p_a11 -= 1
#
#                     # assign image output... (was x_size in the first dimension
#                     # --> needs to be y_size)
#                     image[y_size - 1 - i, j] = ((1 - w_l) * (1 - w_p) * data[l_a00, p_a00] +
#                                                 (1 - w_l) * w_p * data[l_a01, p_a01] +
#                                                 w_l * (1 - w_p) * data[l_a10, p_a10] +
#                                                 w_l * w_p * data[l_a11, p_a11])
#
#                     # nearest neighbour
#                     # nnl = l0
#                     # nnp = p0
#                     # if dp < -0.5 and nnp > 0 and mask[nnl, nnp - 1] == 0:
#                     #     nnp -= 1
#                     # elif dp > 0.5 and nnp < pmax - 1 and mask[nnl, nnp + 1] == 0:
#                     #     nnp += 1
#                     # if dl < -0.5 and nnl > 0 and mask[nnl - 1, nnp] == 0:
#                     #     nnl -= 1
#                     # elif dl > 0.5 and nnl < lmax - 1 and mask[nnl + 1, nnp] == 0:
#                     #     nnl += 1
#                     # image[y_size - 1 - i, j] = data[nnl, nnp]
#
#                     # we found our solution, next
#                     break
#                 else:
#                     # increment...
#                     l0 = int(l0 + dl)
#                     p0 = int(p0 + dp)
#                     if dl > 0 and l0 % chunk_size == chunk_size - 1:
#                         l0 += 2
#                     if dl < 0 and l0 % chunk_size == 0:
#                         l0 -= 2
#     # return the output image
#     print "undefined", undefined
#     return image
#
#
# @cython.boundscheck(False)
# def fast_gradient_search_with_mask_old(np.ndarray[DTYPE_t, ndim=2] data, np.ndarray[DTYPE_t, ndim=2] src_x, np.ndarray[DTYPE_t, ndim=2] src_y, area_extent, size, np.ndarray[np.uint8_t, ndim=2] mask):
#     """Gradient search, simple case variant, with masked data.
#     """
#     # area extent bounds --> area_def.area_extent
#     cdef double x_min, y_min, x_max, y_max
#     x_min, y_min, x_max, y_max = area_extent
#     # change the output size (x_size, y_size) to match area_def.shape:
#     # (lines,pixels)
#     cdef int x_size, y_size
#     y_size, x_size = size
#     # step in x direction (meters); x-output locations (centre of pixel);
#     # repeat for y
#     cdef double x_inc = (x_max - x_min) / x_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_x = np.arange(x_min + x_inc / 2, x_max, x_inc)
#     cdef double y_inc = (y_max - y_min) / y_size
#     cdef np.ndarray[DTYPE_t, ndim = 1] dst_y = np.arange(y_min + y_inc / 2, y_max, y_inc)
#     # output image array --> needs to be (lines, pixels) --> y,x
#     cdef np.ndarray[DTYPE_t, ndim = 2] image = np.zeros([y_size, x_size], dtype=DTYPE)
#     # gradient of output y/x grids (in pixel and line directions)
#     cdef np.ndarray[DTYPE_t, ndim = 2] yp, yl
#     yl, yp = np.gradient(src_y)
#     cdef np.ndarray[DTYPE_t, ndim = 2] xp, xl
#     xl, xp = np.gradient(src_x)
#     # pixel max ---> data is expected in [lines, pixels]
#     cdef int pmax = data.shape[1] - 1
#     cdef int lmax = data.shape[0] - 1
#     # centre of input image - starting point
#     cdef int p0 = pmax / 2
#     cdef int l0 = lmax / 2
#     # intermediate variables:
#     cdef int l_a, l_b, p_a, p_b, nnl, nnp
#     cdef size_t i, j
#     cdef double dx, dy, d, dl, dp, w_l, w_p
#     dl = 0
#     dp = 0
#     # number of iterations
#     cdef int cnt = 0
#     # this was a bit confusing -- "lines" was based on x_size; change this
#     # variable to elements - make it a numpy array (long==numpy int dtype)
#     cdef np.ndarray[size_t, ndim = 1] elements = np.arange(x_size, dtype=np.uintp)
#     for i in range(y_size):
#         # lines.reverse() --> swapped to elements - provide a reverse view of
#         # the array
#         elements = elements[::-1]
#         for j in elements:
#             cnt = 0
#             while True:
#                 cnt += 1
#
#                 # algorithm does not converge.
#                 if cnt > 5:
#                     # TODO: this should be done dynamically or from arg
#                     # distance = np.sqrt((src_y[l0, p0] - dst_y[i]) ** 2 +
#                     #                   (src_x[l0, p0] - dst_x[j]) ** 2)
#                     # if distance < 5000 and mask[l0, p0] != 0:
#                     #    image[y_size - 1 - i, j] = data[l0, p0]
#                     break
#                 # check we are within the input image bounds
#                 if l0 < lmax and l0 >= 0 and p0 < pmax and p0 >= 0:
#                     if mask[l0, p0] != 0:
#                         if dl >= 0:
#                             while l0 < lmax and mask[l0, p0] != 0:
#                                 l0 += 1
#                             if l0 >= lmax:
#                                 l0 = lmax - 1
#                                 break
#                         else:
#                             while l0 >= 0 and mask[l0, p0] != 0:
#                                 l0 -= 1
#                             if l0 < 0:
#                                 l0 = 0
#                                 break
#                 else:
#                     # reset such that we are back in the input image bounds
#                     l0 = max(0, min(lmax - 1, l0))
#                     p0 = max(0, min(pmax - 1, p0))
#                 # step size
#                 dx = dst_x[j] - src_x[l0, p0]
#                 dy = dst_y[i] - src_y[l0, p0]
#
#                 # distance from pixel/line to output location
#                 d = yl[l0, p0] * xp[l0, p0] - yp[l0, p0] * xl[l0, p0]
#                 if d == 0:
#                     image[y_size - 1 - i, j] = data[l0, p0]
#                     break
#                 dp = (yl[l0, p0] * dx - xl[l0, p0] * dy) / d
#                 dl = (xp[l0, p0] * dy - yp[l0, p0] * dx) / d
#                 # check that our distance to an output location is less than 1
#                 # pixel/line
#                 if abs(dp) < 1 and abs(dl) < 1:
#                     # nearest neighbour
#                     # nnl = l0
#                     # if dl < -0.5 and nnl > 0:
#                     #     nnl -= 1
#                     # elif dl > 0.5 and nnl < lmax - 1:
#                     #     nnl += 1
#                     # nnp = p0
#                     # if dp < -0.5 and nnp > 0:
#                     #     nnp -= 1
#                     # elif dp > 0.5 and nnp < pmax - 1:
#                     #     nnp += 1
#                     # image[y_size - 1 - i, j] = data[nnl, nnp]
#                     # bilinear interpolation
#                     if dl < 0:
#                         l_a = max(0, l0 - 1)
#                         l_b = l0
#                         w_l = 1 + dl
#                     else:
#                         l_a = l0
#                         l_b = min(l0 + 1, lmax - 1)
#                         w_l = dl
#                     if dp < 0:
#                         p_a = max(0, p0 - 1)
#                         p_b = p0
#                         w_p = 1 + dp
#                     else:
#                         p_a = p0
#                         p_b = min(p0 + 1, pmax - 1)
#                         w_p = dp
#                     # assign image output... (was x_size in the first dimension
#                     # --> needs to be y_size)
#                     image[y_size - 1 - i, j] = ((1 - w_l) * (1 - w_p) * data[l_a, p_a] +
#                                                 (1 - w_l) * w_p * data[l_a, p_b] +
#                                                 w_l * (1 - w_p) * data[l_b, p_a] +
#                                                 w_l * w_p * data[l_b, p_b])
#                     if mask[l0, p0] != 0:
#                         if dl >= 0:
#                             while l0 < lmax and mask[l0, p0] != 0:
#                                 l0 += 1
#                             if l0 >= lmax:
#                                 l0 = lmax - 1
#                                 break
#                         else:
#                             while l0 >= 0 and mask[l0, p0] != 0:
#                                 l0 -= 1
#                             if l0 < 0:
#                                 l0 = 0
#                                 break
#
#                     image[y_size - 1 - i, j] = data[l0, p0]
#                     # found our solution, next
#                     break
#                 else:
#                     # increment...
#                     l0 = int(l0 + dl)
#                     p0 = int(p0 + dp)
#     # return the output image
#     return image



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'nearest' resampler."""

from typing import Any
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pytest_lazyfixture import lazy_fixture

from pyresample.future.geometry import AreaDefinition, SwathDefinition
from pyresample.future.resamplers import KDTreeNearestXarrayResampler
from pyresample.test.utils import (
    assert_maximum_dask_computes,
    assert_warnings_contain,
    catch_warnings,
)
from pyresample.utils.errors import PerformanceWarning


def _check_common_metadata(data_arr: Any, target_is_area: bool = False) -> None:
    if not isinstance(data_arr, xr.DataArray):
        return
    if not target_is_area:
        return

    for coord_name in ("y", "x"):
        assert coord_name in data_arr.coords
        c_arr = data_arr.coords[coord_name]
        assert c_arr.attrs.get("units") in ("meter", "degrees_north", "degrees_east")
    assert "y" in data_arr.coords
    assert "x" in data_arr.coords


class TestNearestNeighborResampler:
    """Test the KDTreeNearestXarrayResampler class."""

    def test_nearest_swath_1d_mask_to_grid_1n(
            self,
            swath_def_1d_xarray_dask, data_1d_float32_xarray_dask, coord_def_2d_float32_dask):
        """Test 1D swath definition to 2D grid definition; 1 neighbor."""
        resampler = KDTreeNearestXarrayResampler(swath_def_1d_xarray_dask, coord_def_2d_float32_dask)
        res = resampler.resample(data_1d_float32_xarray_dask,
                                 mask_area=data_1d_float32_xarray_dask.isnull(),
                                 radius_of_influence=100000)
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        _check_common_metadata(res, isinstance(coord_def_2d_float32_dask, AreaDefinition))
        actual = res.values
        expected = np.array([
            [1., 2., 2.],
            [1., 2., 2.],
            [1., np.nan, 2.],
            [1., 2., 2.],
        ])
        np.testing.assert_allclose(actual, expected)

    def test_nearest_type_preserve(self, swath_def_1d_xarray_dask, coord_def_2d_float32_dask):
        """Test 1D swath definition to 2D grid definition; 1 neighbor."""
        data = xr.DataArray(da.from_array(np.array([1, 2, 3]),
                                          chunks=5),
                            dims=('my_dim1',))

        resampler = KDTreeNearestXarrayResampler(swath_def_1d_xarray_dask, coord_def_2d_float32_dask)
        res = resampler.resample(data, fill_value=255,
                                 radius_of_influence=100000)
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        _check_common_metadata(res, isinstance(coord_def_2d_float32_dask, AreaDefinition))
        actual = res.values
        expected = np.array([
            [1, 2, 2],
            [1, 2, 2],
            [1, 255, 2],
            [1, 2, 2],
        ])
        np.testing.assert_equal(actual, expected)

    def test_nearest_swath_2d_mask_to_area_1n(self, swath_def_2d_xarray_dask, data_2d_float32_xarray_dask,
                                              area_def_stere_target):
        """Test 2D swath definition to 2D area definition; 1 neighbor."""
        resampler = KDTreeNearestXarrayResampler(
            swath_def_2d_xarray_dask, area_def_stere_target)
        res = resampler.resample(data_2d_float32_xarray_dask, radius_of_influence=50000)
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        _check_common_metadata(res, isinstance(area_def_stere_target, AreaDefinition))
        res = res.values
        cross_sum = float(np.nansum(res))
        expected = 167913.0
        assert cross_sum == expected
        assert res.shape == resampler.target_geo_def.shape

    def test_nearest_area_2d_to_area_1n(self, area_def_stere_source, data_2d_float32_xarray_dask,
                                        area_def_stere_target):
        """Test 2D area definition to 2D area definition; 1 neighbor."""
        resampler = KDTreeNearestXarrayResampler(
            area_def_stere_source, area_def_stere_target)
        with assert_maximum_dask_computes(0):
            resampler.precompute(radius_of_influence=50000)

        with assert_maximum_dask_computes(0):
            res = resampler.resample(data_2d_float32_xarray_dask, radius_of_influence=50000)
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        _check_common_metadata(res, isinstance(area_def_stere_target, AreaDefinition))
        res = res.values
        cross_sum = float(np.nansum(res))
        expected = 303048.0
        assert cross_sum == expected
        assert res.shape == resampler.target_geo_def.shape

    def test_nearest_area_2d_to_area_1n_no_roi(self, area_def_stere_source, data_2d_float32_xarray_dask,
                                               area_def_stere_target):
        """Test 2D area definition to 2D area definition; 1 neighbor, no radius of influence."""
        resampler = KDTreeNearestXarrayResampler(
            area_def_stere_source, area_def_stere_target)
        resampler.precompute()

        res = resampler.resample(data_2d_float32_xarray_dask)
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        _check_common_metadata(res, isinstance(area_def_stere_target, AreaDefinition))
        res = res.values
        cross_sum = float(np.nansum(res))
        expected = 952386.0
        assert cross_sum == expected
        assert res.shape == resampler.target_geo_def.shape

    def test_nearest_area_2d_to_area_1n_no_roi_no_geocentric(
            self,
            area_def_stere_source,
            data_2d_float32_xarray_dask,
            area_def_stere_target):
        # pretend the resolutions can't be determined
        with mock.patch.object(area_def_stere_source, 'geocentric_resolution') as sgr, \
                mock.patch.object(area_def_stere_target, 'geocentric_resolution') as dgr:
            sgr.side_effect = RuntimeError
            dgr.side_effect = RuntimeError
            resampler = KDTreeNearestXarrayResampler(
                area_def_stere_source, area_def_stere_target)
            res = resampler.resample(data_2d_float32_xarray_dask)
            assert isinstance(res, xr.DataArray)
            assert isinstance(res.data, da.Array)
            _check_common_metadata(res, isinstance(area_def_stere_target, AreaDefinition))
            res = res.values
            cross_sum = np.nansum(res)
            expected = 20666.0
            assert cross_sum == expected
            assert res.shape == resampler.target_geo_def.shape

    @pytest.mark.parametrize("input_data", [
        lazy_fixture("data_2d_float32_numpy"),
        lazy_fixture("data_2d_float32_dask"),
        lazy_fixture("data_2d_float32_xarray_numpy"),
    ])
    def test_object_type_with_warnings(
            self,
            area_def_stere_source,
            area_def_stere_target,
            input_data):
        """Test that providing certain input data causes a warning."""
        resampler = KDTreeNearestXarrayResampler(area_def_stere_source, area_def_stere_target)
        with catch_warnings(PerformanceWarning) as w, assert_maximum_dask_computes(1):
            res = resampler.resample(input_data)
            assert type(res) is type(input_data)
        _check_common_metadata(res, isinstance(area_def_stere_target, AreaDefinition))
        is_data_arr_dask = isinstance(input_data, xr.DataArray) and isinstance(input_data.data, da.Array)
        is_dask_based = isinstance(input_data, da.Array) or is_data_arr_dask
        if is_dask_based:
            assert not w
        else:
            assert_warnings_contain(w, "will be converted")
        res = np.array(res)
        cross_sum = np.nansum(res)
        expected = 952386.0  # same as 'test_nearest_area_2d_to_area_1n_no_roi'
        assert cross_sum == expected
        assert res.shape == resampler.target_geo_def.shape

    def test_nearest_area_2d_to_area_1n_3d_data(self, area_def_stere_source, data_3d_float32_xarray_dask,
                                                area_def_stere_target):
        """Test 2D area definition to 2D area definition; 1 neighbor, 3d data."""
        resampler = KDTreeNearestXarrayResampler(
            area_def_stere_source, area_def_stere_target)
        resampler.precompute(radius_of_influence=50000)

        res = resampler.resample(data_3d_float32_xarray_dask, radius_of_influence=50000)
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert list(res.coords['bands']) == ['r', 'g', 'b']
        _check_common_metadata(res, isinstance(area_def_stere_target, AreaDefinition))
        res = res.values
        cross_sum = float(np.nansum(res))
        expected = 909144.0
        assert cross_sum == expected
        assert res.shape[:2] == resampler.target_geo_def.shape


class TestInvalidUsageNearestNeighborResampler:
    """Test the resampler being given input that should raise an error.

    If a case here is removed because functionality is added to the resampler
    then a working case should be added above.

    """

    @pytest.mark.parametrize(
        "input_data",
        [
            lazy_fixture("data_2d_float32_xarray_dask"),
            lazy_fixture("data_3d_float32_xarray_dask"),
        ]
    )
    def test_mismatch_geo_data_dims(
            self,
            area_def_stere_source,
            area_def_stere_target,
            input_data,
    ):
        resampler = KDTreeNearestXarrayResampler(area_def_stere_source, area_def_stere_target)
        data = input_data.rename({'y': 'my_dim_y', 'x': 'my_dim_x'})
        with pytest.raises(ValueError, match='.*dimensions do not match.*'):
            resampler.resample(data)

    def test_mismatch_geo_data_dims_swath(
            self,
            swath_def_2d_xarray_dask,
            area_def_stere_target,
            data_2d_float32_xarray_dask):
        new_swath_def = SwathDefinition(
            swath_def_2d_xarray_dask.lons.rename({'y': 'my_dim_y', 'x': 'my_dim_x'}),
            swath_def_2d_xarray_dask.lats.rename({'y': 'my_dim_y', 'x': 'my_dim_x'})
        )
        resampler = KDTreeNearestXarrayResampler(new_swath_def, area_def_stere_target)
        with pytest.raises(ValueError, match='.*dimensions do not match.*'):
            resampler.resample(data_2d_float32_xarray_dask)

    @pytest.mark.parametrize(
        "src_geom",
        [
            lazy_fixture("area_def_stere_source"),
            lazy_fixture("swath_def_2d_xarray_dask")
        ]
    )
    @pytest.mark.parametrize(
        ("match", "call_precompute"),
        [
            (".*data.*shape.*", False),
            (".*'mask'.*shape.*", True),
        ]
    )
    def test_inconsistent_input_shapes(self, src_geom, match, call_precompute,
                                       area_def_stere_target, data_2d_float32_xarray_dask):
        """Test that geometry and data of the same size but different size still error."""
        # transpose the source geometries
        if isinstance(src_geom, AreaDefinition):
            src_geom = AreaDefinition(
                src_geom.area_id, src_geom.description, src_geom.proj_id,
                src_geom.crs, src_geom.height, src_geom.width, src_geom.area_extent,
            )
        else:
            src_geom = SwathDefinition(
                src_geom.lons.T.rename({'y': 'x', 'x': 'y'}),
                src_geom.lats.T.rename({'y': 'x', 'x': 'y'}),
            )
        resampler = KDTreeNearestXarrayResampler(src_geom, area_def_stere_target)
        with pytest.raises(ValueError, match=match):
            if call_precompute:
                resampler.precompute(mask=data_2d_float32_xarray_dask.notnull())
            else:
                resampler.resample(data_2d_float32_xarray_dask)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Definitions for geographic regions."""

from typing import Union

from .area import AreaDefinition, DynamicAreaDefinition  # noqa
from .swath import CoordinateDefinition, SwathDefinition  # noqa

StaticGeometry = Union[SwathDefinition, CoordinateDefinition, AreaDefinition]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2019

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>

# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the gradien search resampling."""

import unittest
from unittest import mock

import dask.array as da
import numpy as np
import xarray as xr

from pyresample.geometry import AreaDefinition, SwathDefinition


class TestGradientResampler(unittest.TestCase):
    """Test case for the gradient resampling."""

    def setUp(self):
        """Set up the test case."""
        from pyresample.gradient import GradientSearchResampler
        self.src_area = AreaDefinition('dst', 'dst area', None,
                                       {'ellps': 'WGS84', 'h': '35785831', 'proj': 'geos'},
                                       100, 100,
                                       (5550000.0, 5550000.0, -5550000.0, -5550000.0))
        self.src_swath = SwathDefinition(*self.src_area.get_lonlats())
        self.dst_area = AreaDefinition('euro40', 'euro40', None,
                                       {'proj': 'stere', 'lon_0': 14.0,
                                        'lat_0': 90.0, 'lat_ts': 60.0,
                                        'ellps': 'bessel'},
                                       102, 102,
                                       (-2717181.7304994687, -5571048.14031214,
                                        1378818.2695005313, -1475048.1403121399))

        self.resampler = GradientSearchResampler(self.src_area, self.dst_area)
        self.swath_resampler = GradientSearchResampler(self.src_swath,
                                                       self.dst_area)

    def test_get_projection_coordinates_area_to_area(self):
        """Check that the coordinates are initialized, for area -> area."""
        assert self.resampler.prj is None
        self.resampler._get_projection_coordinates((10, 10))
        cdst_x = self.resampler.dst_x.compute()
        cdst_y = self.resampler.dst_y.compute()
        assert np.allclose(np.min(cdst_x), -2022632.1675016289)
        assert np.allclose(np.max(cdst_x), 2196052.591296284)
        assert np.allclose(np.min(cdst_y), 3517933.413092212)
        assert np.allclose(np.max(cdst_y), 5387038.893400168)
        assert self.resampler.use_input_coords
        assert self.resampler.prj is not None

    def test_get_projection_coordinates_swath_to_area(self):
        """Check that the coordinates are initialized, for swath -> area."""
        assert self.swath_resampler.prj is None
        self.swath_resampler._get_projection_coordinates((10, 10))
        cdst_x = self.swath_resampler.dst_x.compute()
        cdst_y = self.swath_resampler.dst_y.compute()
        assert np.allclose(np.min(cdst_x), -2697103.29912692)
        assert np.allclose(np.max(cdst_x), 1358739.8381279823)
        assert np.allclose(np.min(cdst_y), -5550969.708939591)
        assert np.allclose(np.max(cdst_y), -1495126.5716846888)
        assert self.swath_resampler.use_input_coords is False
        assert self.swath_resampler.prj is not None

    def test_get_gradients(self):
        """Test that coordinate gradients are computed correctly."""
        self.resampler._get_projection_coordinates((10, 10))
        assert self.resampler.src_gradient_xl is None
        self.resampler._get_gradients()
        assert self.resampler.src_gradient_xl.compute().max() == 0.0
        assert self.resampler.src_gradient_xp.compute().max() == -111000.0
        assert self.resampler.src_gradient_yl.compute().max() == 111000.0
        assert self.resampler.src_gradient_yp.compute().max() == 0.0

    def test_get_chunk_mappings(self):
        """Test that chunk overlap, and source and target slices are correct."""
        chunks = (10, 10)
        num_chunks = np.product(chunks)
        self.resampler._get_projection_coordinates(chunks)
        self.resampler._get_gradients()
        assert self.resampler.coverage_status is None
        self.resampler.get_chunk_mappings()
        # 8 source chunks overlap the target area
        covered_src_chunks = np.array([38, 39, 48, 49, 58, 59, 68, 69])
        res = np.where(self.resampler.coverage_status)[0]
        assert np.all(res == covered_src_chunks)
        # All *num_chunks* should have values in the lists
        assert len(self.resampler.coverage_status) == num_chunks
        assert len(self.resampler.src_slices) == num_chunks
        assert len(self.resampler.dst_slices) == num_chunks
        assert len(self.resampler.dst_mosaic_locations) == num_chunks
        # There's only one output chunk, and the covered source chunks
        # should have destination locations of (0, 0)
        res = np.array(self.resampler.dst_mosaic_locations)[covered_src_chunks]
        assert all([all(loc == (0, 0)) for loc in list(res)])

    def test_get_src_poly_area(self):
        """Test defining source chunk polygon for AreaDefinition."""
        chunks = (10, 10)
        self.resampler._get_projection_coordinates(chunks)
        self.resampler._get_gradients()
        poly = self.resampler._get_src_poly(0, 40, 0, 40)
        assert np.allclose(poly.area, 12365358458842.43)

    def test_get_src_poly_swath(self):
        """Test defining source chunk polygon for SwathDefinition."""
        chunks = (10, 10)
        self.swath_resampler._get_projection_coordinates(chunks)
        self.swath_resampler._get_gradients()
        # Swath area defs can't be sliced, so False is returned
        poly = self.swath_resampler._get_src_poly(0, 40, 0, 40)
        assert poly is False

    @mock.patch('pyresample.gradient.get_polygon')
    def test_get_dst_poly(self, get_polygon):
        """Test defining destination chunk polygon."""
        chunks = (10, 10)
        self.resampler._get_projection_coordinates(chunks)
        self.resampler._get_gradients()
        # First call should make a call to get_polygon()
        self.resampler._get_dst_poly('idx1', 0, 10, 0, 10)
        assert get_polygon.call_count == 1
        assert 'idx1' in self.resampler.dst_polys
        # The second call to the same index should come from cache
        self.resampler._get_dst_poly('idx1', 0, 10, 0, 10)
        assert get_polygon.call_count == 1

        # Swath defs raise AttributeError, and False is returned
        get_polygon.side_effect = AttributeError
        self.resampler._get_dst_poly('idx2', 0, 10, 0, 10)
        assert self.resampler.dst_polys['idx2'] is False

    def test_filter_data(self):
        """Test filtering chunks that do not overlap."""
        chunks = (10, 10)
        self.resampler._get_projection_coordinates(chunks)
        self.resampler._get_gradients()
        self.resampler.get_chunk_mappings()

        # Basic filtering.  There should be 8 dask arrays that each
        # have a shape of (10, 10)
        res = self.resampler._filter_data(self.resampler.src_x)
        valid = [itm for itm in res if itm is not None]
        assert len(valid) == 8
        shapes = [arr.shape for arr in valid]
        for shp in shapes:
            assert shp == (10, 10)

        # Destination x/y coordinate array filtering.  Again, 8 dask
        # arrays each with shape (102, 102)
        res = self.resampler._filter_data(self.resampler.dst_x, is_src=False)
        valid = [itm for itm in res if itm is not None]
        assert len(valid) == 8
        shapes = [arr.shape for arr in valid]
        for shp in shapes:
            assert shp == (102, 102)

        # Add a dimension to the given dataset
        data = da.random.random(self.src_area.shape)
        res = self.resampler._filter_data(data, add_dim=True)
        valid = [itm for itm in res if itm is not None]
        assert len(valid) == 8
        shapes = [arr.shape for arr in valid]
        for shp in shapes:
            assert shp == (1, 10, 10)

        # 1D and 3+D should raise NotImplementedError
        data = da.random.random((3,))
        try:
            res = self.resampler._filter_data(data, add_dim=True)
            raise IndexError
        except NotImplementedError:
            pass
        data = da.random.random((3, 3, 3, 3))
        try:
            res = self.resampler._filter_data(data, add_dim=True)
            raise IndexError
        except NotImplementedError:
            pass

    def test_resample_area_to_area_2d(self):
        """Resample area to area, 2d."""
        data = xr.DataArray(da.ones(self.src_area.shape, dtype=np.float64),
                            dims=['y', 'x'])
        res = self.resampler.compute(
            data, method='bil').compute(scheduler='single-threaded')
        assert res.shape == self.dst_area.shape
        assert np.allclose(res, 1)

    def test_resample_area_to_area_2d_fill_value(self):
        """Resample area to area, 2d, use fill value."""
        data = xr.DataArray(da.full(self.src_area.shape, np.nan,
                                    dtype=np.float64), dims=['y', 'x'])
        res = self.resampler.compute(
            data, method='bil',
            fill_value=2.0).compute(scheduler='single-threaded')
        assert res.shape == self.dst_area.shape
        assert np.allclose(res, 2.0)

    def test_resample_area_to_area_3d(self):
        """Resample area to area, 3d."""
        data = xr.DataArray(da.ones((3, ) + self.src_area.shape,
                                    dtype=np.float64) *
                            np.array([1, 2, 3])[:, np.newaxis, np.newaxis],
                            dims=['bands', 'y', 'x'])
        res = self.resampler.compute(
            data, method='bil').compute(scheduler='single-threaded')
        assert res.shape == (3, ) + self.dst_area.shape
        assert np.allclose(res[0, :, :], 1.0)
        assert np.allclose(res[1, :, :], 2.0)
        assert np.allclose(res[2, :, :], 3.0)

    def test_resample_swath_to_area_2d(self):
        """Resample swath to area, 2d."""
        data = xr.DataArray(da.ones(self.src_swath.shape, dtype=np.float64),
                            dims=['y', 'x'])
        res = self.swath_resampler.compute(
            data, method='bil').compute(scheduler='single-threaded')
        assert res.shape == self.dst_area.shape
        assert not np.all(np.isnan(res))

    def test_resample_swath_to_area_3d(self):
        """Resample area to area, 3d."""
        data = xr.DataArray(da.ones((3, ) + self.src_swath.shape,
                                    dtype=np.float64) *
                            np.array([1, 2, 3])[:, np.newaxis, np.newaxis],
                            dims=['bands', 'y', 'x'])
        res = self.swath_resampler.compute(
            data, method='bil').compute(scheduler='single-threaded')
        assert res.shape == (3, ) + self.dst_area.shape
        for i in range(res.shape[0]):
            arr = np.ravel(res[i, :, :])
            assert np.allclose(arr[np.isfinite(arr)], float(i + 1))


def test_check_overlap():
    """Test overlap check returning correct results."""
    from shapely.geometry import Polygon

    from pyresample.gradient import check_overlap

    # If either of the polygons is False, True is returned
    assert check_overlap(False, 3) is True
    assert check_overlap('eggs', False) is True
    assert check_overlap(False, False) is True

    # If either the polygons is None, False is returned
    assert check_overlap(None, 'bacon') is False
    assert check_overlap('spam', None) is False
    assert check_overlap(None, None) is False

    # If the polygons overlap, True is returned
    poly1 = Polygon(((0, 0), (0, 1), (1, 1), (1, 0)))
    poly2 = Polygon(((-1, -1), (-1, 1), (1, 1), (1, -1)))
    assert check_overlap(poly1, poly2) is True

    # If the polygons do not overlap, False is returned
    poly2 = Polygon(((5, 5), (6, 5), (6, 6), (5, 6)))
    assert check_overlap(poly1, poly2) is False


@mock.patch('pyresample.gradient.get_geostationary_bounding_box')
def test_get_border_lonlats(get_geostationary_bounding_box):
    """Test that correct methods are called in get_border_lonlats()."""
    from pyresample.gradient import get_border_lonlats
    geo_def = mock.MagicMock(proj_dict={'proj': 'geos'})
    get_geostationary_bounding_box.return_value = 1, 2
    res = get_border_lonlats(geo_def)
    assert res == (1, 2)
    get_geostationary_bounding_box.assert_called_with(geo_def, 3600)
    geo_def.get_boundary_lonlats.assert_not_called()

    lon_sides = mock.MagicMock(side1=np.array([1]), side2=np.array([2]),
                               side3=np.array([3]), side4=np.array([4]))
    lat_sides = mock.MagicMock(side1=np.array([1]), side2=np.array([2]),
                               side3=np.array([3]), side4=np.array([4]))
    geo_def = mock.MagicMock()
    geo_def.get_boundary_lonlats.return_value = lon_sides, lat_sides
    lon_b, lat_b = get_border_lonlats(geo_def)
    assert np.all(lon_b == np.array([1, 2, 3, 4]))
    assert np.all(lat_b == np.array([1, 2, 3, 4]))


@mock.patch('pyresample.gradient.Polygon')
@mock.patch('pyresample.gradient.get_border_lonlats')
def test_get_polygon(get_border_lonlats, Polygon):
    """Test polygon creation."""
    from pyresample.gradient import get_polygon

    # Valid polygon
    get_border_lonlats.return_value = (1, 2)
    geo_def = mock.MagicMock()
    prj = mock.MagicMock()
    x_borders = [0, 0, 1, 1]
    y_borders = [0, 1, 1, 0]
    boundary = [(0, 0), (0, 1), (1, 1), (1, 0)]
    prj.return_value = (x_borders, y_borders)
    poly = mock.MagicMock(area=2.0)
    Polygon.return_value = poly
    res = get_polygon(prj, geo_def)
    get_border_lonlats.assert_called_with(geo_def)
    prj.assert_called_with(1, 2)
    Polygon.assert_called_with(boundary)
    assert res is poly

    # Some border points are invalid, those should have been removed
    x_borders = [np.inf, 0, 0, 0, 1, np.nan, 2]
    y_borders = [-1, 0, np.nan, 1, 1, np.nan, -1]
    boundary = [(0, 0), (0, 1), (1, 1), (2, -1)]
    prj.return_value = (x_borders, y_borders)
    res = get_polygon(prj, geo_def)
    Polygon.assert_called_with(boundary)
    assert res is poly

    # Polygon area is NaN
    poly.area = np.nan
    res = get_polygon(prj, geo_def)
    assert res is None

    # Polygon area is 0.0
    poly.area = 0.0
    res = get_polygon(prj, geo_def)
    assert res is None


@mock.patch('pyresample.gradient.one_step_gradient_search')
def test_gradient_resample_data(one_step_gradient_search):
    """Test that one_step_gradient_search() is called with proper array shapes."""
    from pyresample.gradient import _gradient_resample_data

    ndim_3 = np.zeros((3, 3, 4))
    ndim_2a = np.zeros((3, 4))
    ndim_2b = np.zeros((8, 10))

    # One of the source arrays has wrong shape
    try:
        _ = _gradient_resample_data(ndim_3, ndim_2a, ndim_2b, ndim_2a, ndim_2a,
                                    ndim_2a, ndim_2a, ndim_2b, ndim_2b)
        raise IndexError
    except AssertionError:
        pass
    one_step_gradient_search.assert_not_called()

    # Data array has wrong shape
    try:
        _ = _gradient_resample_data(ndim_2a, ndim_2a, ndim_2a, ndim_2a, ndim_2a,
                                    ndim_2a, ndim_2a, ndim_2b, ndim_2b)
        raise IndexError
    except AssertionError:
        pass
    one_step_gradient_search.assert_not_called()

    # The destination x and y arrays have different shapes
    try:
        _ = _gradient_resample_data(ndim_3, ndim_2a, ndim_2a, ndim_2a, ndim_2a,
                                    ndim_2a, ndim_2a, ndim_2b, ndim_2a)
        raise IndexError
    except AssertionError:
        pass
    one_step_gradient_search.assert_not_called()

    # Correct shapes are given
    _ = _gradient_resample_data(ndim_3, ndim_2a, ndim_2a, ndim_2a, ndim_2a,
                                ndim_2a, ndim_2a, ndim_2b, ndim_2b)
    one_step_gradient_search.assert_called_once()


@mock.patch('pyresample.gradient.dask.delayed')
@mock.patch('pyresample.gradient._concatenate_chunks')
@mock.patch('pyresample.gradient.da')
def test_parallel_gradient_search(dask_da, _concatenate_chunks, delayed):
    """Test calling parallel_gradient_search()."""
    from pyresample.gradient import parallel_gradient_search

    def mock_cc(chunks):
        """Return the input."""
        return chunks

    _concatenate_chunks.side_effect = mock_cc

    # Mismatch in number of bands raises ValueError
    data = [np.zeros((1, 5, 5)), np.zeros((2, 5, 5))]
    try:
        parallel_gradient_search(data, None, None, None, None,
                                 None, None, None, None, None, None)
        raise
    except ValueError:
        pass

    data = [np.zeros((1, 5, 4)), np.ones((1, 5, 4)), None, None]
    src_x, src_y = [1, 2, 3, 4], [4, 5, 6, 4]
    # dst_x is used to check the target area shape, so needs "valid"
    # data.  The last values shouldn't matter as data[-2:] are None
    # and should be skipped.
    dst_x = [np.zeros((5, 5)), np.zeros((5, 5)), 'foo', 'bar']
    dst_y = [1, 2, 3, 4]
    src_gradient_xl, src_gradient_xp = [1, 2, None, None], [1, 2, None, None]
    src_gradient_yl, src_gradient_yp = [1, 2, None, None], [1, 2, None, None]
    # Destination slices are used only for padding, so the first two
    # None values shouldn't raise errors
    dst_slices = [None, None, [1, 2, 1, 3], [1, 3, 1, 4]]
    # The first two chunks have the same target location, same for the two last
    dst_mosaic_locations = [(0, 0), (0, 0), (0, 1), (0, 1)]

    res = parallel_gradient_search(data, src_x, src_y, dst_x, dst_y,
                                   src_gradient_xl, src_gradient_xp,
                                   src_gradient_yl, src_gradient_yp,
                                   dst_mosaic_locations, dst_slices,
                                   method='foo')
    assert len(res[(0, 0)]) == 2
    # The second padding shouldn't be in the chunks[(0, 1)] list
    assert len(res[(0, 1)]) == 1
    _concatenate_chunks.assert_called_with(res)
    # Two padding arrays
    assert dask_da.full.call_count == 2
    assert mock.call((1, 1, 2), np.nan) in dask_da.full.mock_calls
    assert mock.call((1, 2, 3), np.nan) in dask_da.full.mock_calls
    # Two resample calls
    assert dask_da.from_delayed.call_count == 2
    # The _gradient_resample_data() function has been delayed twice
    assert '_gradient_resample_data' in str(delayed.mock_calls[0])
    assert '_gradient_resample_data' in str(delayed.mock_calls[2])
    assert str(mock.call()(data[0],
                           src_x[0], src_y[0],
                           src_gradient_xl[0], src_gradient_xp[0],
                           src_gradient_yl[0], src_gradient_yp[0],
                           dst_x[0], dst_y[0],
                           method='foo')) == str(delayed.mock_calls[1])
    assert str(mock.call()(data[1],
                           src_x[1], src_y[1],
                           src_gradient_xl[1], src_gradient_xp[1],
                           src_gradient_yl[1], src_gradient_yp[1],
                           dst_x[1], dst_y[1],
                           method='foo')) == str(delayed.mock_calls[3])


def test_concatenate_chunks():
    """Test chunk concatenation for correct results."""
    from pyresample.gradient import _concatenate_chunks

    # 1-band image
    chunks = {(0, 0): [np.ones((1, 5, 4)), np.zeros((1, 5, 4))],
              (1, 0): [np.zeros((1, 5, 2))],
              (1, 1): [np.full((1, 3, 2), 0.5)],
              (0, 1): [np.full((1, 3, 4), -1)]}
    res = _concatenate_chunks(chunks).compute(scheduler='single-threaded')
    assert np.all(res[:5, :4] == 1.0)
    assert np.all(res[:5, 4:] == 0.0)
    assert np.all(res[5:, :4] == -1.0)
    assert np.all(res[5:, 4:] == 0.5)
    assert res.shape == (8, 6)

    # 3-band image
    chunks = {(0, 0): [np.ones((3, 5, 4)), np.zeros((3, 5, 4))],
              (1, 0): [np.zeros((3, 5, 2))],
              (1, 1): [np.full((3, 3, 2), 0.5)],
              (0, 1): [np.full((3, 3, 4), -1)]}
    res = _concatenate_chunks(chunks).compute(scheduler='single-threaded')
    assert np.all(res[:, :5, :4] == 1.0)
    assert np.all(res[:, :5, 4:] == 0.0)
    assert np.all(res[:, 5:, :4] == -1.0)
    assert np.all(res[:, 5:, 4:] == 0.5)
    assert res.shape == (3, 8, 6)


@mock.patch('pyresample.gradient.da')
def test_concatenate_chunks_stack_calls(dask_da):
    """Test that stacking is called the correct times in chunk concatenation."""
    from pyresample.gradient import _concatenate_chunks

    chunks = {(0, 0): [np.ones((1, 5, 4)), np.zeros((1, 5, 4))],
              (1, 0): [np.zeros((1, 5, 2))],
              (1, 1): [np.full((1, 3, 2), 0.5)],
              (0, 1): [np.full((1, 3, 4), -1)]}
    _ = _concatenate_chunks(chunks)
    dask_da.stack.assert_called_once_with(chunks[(0, 0)], axis=-1)
    dask_da.nanmax.assert_called_once()
    assert 'axis=2' in str(dask_da.concatenate.mock_calls[-2])
    assert 'squeeze' in str(dask_da.concatenate.mock_calls[-1])
